# distutils: libraries = NTL_LIBRARIES gmp m
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++

# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#       Copyright (C) 2007 Martin Albrecht <malb@informatik.uni-bremen.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.ext.cplusplus cimport ccrepr

include 'misc.pxi'
include 'decl.pxi'

from cpython.object cimport Py_EQ, Py_NE
from sage.libs.ntl.ntl_ZZ cimport ntl_ZZ
from sage.libs.ntl.ntl_GF2 cimport ntl_GF2
from sage.libs.ntl.ntl_GF2X cimport ntl_GF2X
from sage.libs.ntl.ntl_GF2EContext cimport ntl_GF2EContext_class
from sage.libs.ntl.ntl_GF2EContext import ntl_GF2EContext
from sage.libs.ntl.ntl_ZZ import unpickle_class_args
from sage.misc.randstate cimport current_randstate


##############################################################################
#
# ntl_GF2E: GF(2**x) via NTL
#
# AUTHORS:
#  - Martin Albrecht <malb@informatik.uni-bremen.de>
#    2006-01: initial version (based on code by William Stein)
#  - Martin Albrecht <malb@informatik.uni-bremen.de>
#    2007-10: adapted to new conventions
#
##############################################################################

def ntl_GF2E_random(ntl_GF2EContext_class ctx):
    """
    Return a random element from GF2E modulo the current modulus.

    INPUT:

    - ``ctx`` -- the GF2E context for which a random element should be created

    EXAMPLES::

        sage: ctx = ntl.GF2EContext([1,1,0,1,1,0,0,0,1])
        sage: ntl.GF2E_random(ctx).modulus_context() is ctx
        True
    """
    current_randstate().set_seed_ntl(False)

    cdef ntl_GF2E r
    ctx.restore_c()
    r = ntl_GF2E.__new__(ntl_GF2E)
    r.c = ctx
    r.x = GF2E_random()
    return r


cdef class ntl_GF2E():
    r"""
    The :class:`GF2E` represents a finite extension field over GF(2)
    using NTL. Elements are represented as polynomials over GF(2)
    modulo a modulus.

    This modulus must be set by creating a GF2EContext first and pass
    that context to the constructor of all elements.
    """

    def __init__(self, x=None, modulus=None):
        """
        Construct a new finite field element in GF(2**x).

        If you pass a string to the constructor please note that byte
        sequences and the hexadecimal notation are Little Endian in
        NTL.  So e.g. '[0 1]' == '0x2' == x.

        INPUT:

        - ``x`` -- value to be assigned to this element. Same types as
          ``ntl.GF2X()`` are accepted.
        - ``modulus`` -- the context/modulus of the field

        OUTPUT: a new ``ntl.GF2E`` element

        EXAMPLES::

            sage: k.<a> = GF(2^8)
            sage: e = ntl.GF2E(a,k); e
            [0 1]
            sage: ctx = e.modulus_context()
            sage: ntl.GF2E('0x1c', ctx)
            [1 0 0 0 0 0 1 1]
            sage: ntl.GF2E('[1 0 1 0]', ctx)
            [1 0 1]
            sage: ntl.GF2E([1,0,1,0], ctx)
            [1 0 1]
            sage: ntl.GF2E(ntl.GF2(1),ctx)
            [1]
        """
        if modulus is None:
            raise ValueError("You must specify a modulus when creating a GF2E.")

        cdef ntl_GF2X _x

        if isinstance(x, ntl_GF2X):
            GF2E_conv_GF2X(self.x, (<ntl_GF2X>x).x)

        elif isinstance(x, int):
            GF2E_conv_long(self.x, x)

        elif isinstance(x, ntl_ZZ):
            GF2E_conv_ZZ(self.x, (<ntl_ZZ>x).x)

        elif isinstance(x, ntl_GF2):
            GF2E_conv_GF2(self.x, (<ntl_GF2>x).x)
        else:
            _x = ntl_GF2X(x)
            GF2E_conv_GF2X(self.x, _x.x)

    def __cinit__(self, x=None, modulus=None):
        #################### WARNING ###################
        ## Before creating a GF2E, you must create a  ##
        ## GF2EContext, and restore it.  In Python,   ##
        ## the error checking in __init__ will prevent##
        ## you from constructing an ntl_GF2E          ##
        ## inappropriately.  However, from Cython, you##
        ## could do r = ntl_GF2E.__new__(ntl_GF2E) without
        ## first restoring a GF2EContext, which could ##
        ## have unfortunate consequences.  See _new   ##
        ## defined below for an example of the right  ##
        ## way to short-circuit __init__ (or just call##
        ## _new in your own code).                    ##
        ################################################
        if modulus is None:
            return
        if isinstance(modulus, ntl_GF2EContext_class):
            self.c = <ntl_GF2EContext_class>modulus
            self.c.restore_c()
        else:
            self.c = <ntl_GF2EContext_class>ntl_GF2EContext(modulus)
            self.c.restore_c()

    cdef ntl_GF2E _new(self):
        cdef ntl_GF2E r
        self.c.restore_c()
        r = ntl_GF2E.__new__(ntl_GF2E)
        r.c = self.c
        return r

    def __reduce__(self):
        """
        EXAMPLES::

            sage: ctx = ntl.GF2EContext( ntl.GF2X([1,1,0,1,1,0,0,0,1]) )
            sage: a = ntl.GF2E(ntl.ZZ_pX([1,1,3],2), ctx)
            sage: loads(dumps(a)) == a
            True
        """
        return unpickle_class_args, (ntl_GF2E, (str(self), self.modulus_context()))

    def modulus_context(self):
        """
        Return the structure that holds the underlying NTL GF2E modulus.

        EXAMPLES::

            sage: ctx = ntl.GF2EContext( ntl.GF2X([1,1,0,1,1,0,0,0,1]) )
            sage: a = ntl.GF2E(ntl.ZZ_pX([1,1,3],2), ctx)
            sage: cty = a.modulus_context(); cty
            NTL modulus [1 1 0 1 1 0 0 0 1]
            sage: ctx == cty
            True
        """
        return self.c

    def __repr__(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: ntl.GF2E([1,1,0,1], ctx) # indirect doctest
            [1 1 0 1]
        """
        self.c.restore_c()
        return ccrepr(self.x)

    def __copy__(self):
        """
        Return a copy of ``self``.

        EXAMPLES::

            sage: x = ntl.GF2E([0,1,1],GF(2^4,'a'))
            sage: y = copy(x)
            sage: x == y
            True
            sage: x is y
            False
        """
        cdef ntl_GF2E r = self._new()
        r.x = self.x
        return r

    def __mul__(ntl_GF2E self, other):
        """
        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: x = ntl.GF2E([1,0,1,0,1], ctx) ; y = ntl.GF2E([1,1,0,1,1], ctx)
            sage: x*y ## indirect doctest
            [0 0 1 1 1 0 1 1]
        """
        cdef ntl_GF2E r
        if not isinstance(other, ntl_GF2E):
            other = ntl_GF2E(other,self.c)
        elif self.c is not (<ntl_GF2E>other).c:
            raise ValueError("You cannot perform arithmetic with elements in different fields.")
        r = self._new()
        GF2E_mul(r.x, self.x, (<ntl_GF2E>other).x)
        return r

    def __sub__(ntl_GF2E self, other):
        """
        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: x = ntl.GF2E([1,0,1,0,1], ctx) ; y = ntl.GF2E([1,1,0,1,1], ctx)
            sage: x - y ## indirect doctest
            [0 1 1 1]
        """
        cdef ntl_GF2E r
        if not isinstance(other, ntl_GF2E):
            other = ntl_GF2E(other,self.c)
        elif self.c is not (<ntl_GF2E>other).c:
            raise ValueError("You cannot perform arithmetic with elements in different fields.")
        r = self._new()
        GF2E_sub(r.x, self.x, (<ntl_GF2E>other).x)
        return r

    def __add__(ntl_GF2E self, other):
        """
        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: x = ntl.GF2E([1,0,1,0,1], ctx) ; y = ntl.GF2E([1,1,0,1,1], ctx)
            sage: x+y ## indirect doctest
            [0 1 1 1]
        """
        cdef ntl_GF2E r
        if not isinstance(other, ntl_GF2E):
            other = ntl_GF2E(other,self.c)
        elif self.c is not (<ntl_GF2E>other).c:
            raise ValueError("You cannot perform arithmetic with elements in different fields.")
        r = self._new()
        GF2E_add(r.x, self.x, (<ntl_GF2E>other).x)
        return r

    def __truediv__(ntl_GF2E self, other):
        """
        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: x = ntl.GF2E([1,0,1,0,1], ctx) ; y = ntl.GF2E([1,1,0,1,1], ctx)
            sage: x/y ## indirect doctest
            [1 0 1 0 0 1 0 1]
        """
        cdef ntl_GF2E r
        if not isinstance(other, ntl_GF2E):
            other = ntl_GF2E(other, self.c)
        elif self.c is not (<ntl_GF2E>other).c:
            raise ValueError("You cannot perform arithmetic with elements in different fields.")
        r = self._new()
        GF2E_div(r.x, self.x, (<ntl_GF2E>other).x)
        return r

    def __neg__(ntl_GF2E self):
        """
        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: x = ntl.GF2E([1,0,1,0,1], ctx)
            sage: -x ## indirect doctest
            [1 0 1 0 1]
        """
        cdef ntl_GF2E r = self._new()
        r.x = self.x
        return r

    def __pow__(ntl_GF2E self, long e, ignored):
        """
        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: x = ntl.GF2E([1,0,1,0,1], ctx)
            sage: x**2 ## indirect doctest
            [0 1 0 1]
        """
        cdef ntl_GF2E r = self._new()
        GF2E_power(r.x, self.x, e)
        return r

    def __richcmp__(ntl_GF2E self, other, int op):
        r"""
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: x = ntl.GF2E([1,0,1,0,1], ctx) ; y = ntl.GF2E([1,1,0,1,1], ctx)
            sage: x == x
            True
            sage: x == y
            False
            sage: ntl.GF2E(0,ctx) == 0
            True
            sage: a = ntl.GF2E([0,1],GF(2^2,'a'))
            sage: a == x
            False
        """
        self.c.restore_c()

        if op != Py_EQ and op != Py_NE:
            raise TypeError("elements of GF(2^e) are not ordered")

        cdef ntl_GF2E b
        try:
            b = <ntl_GF2E?>other
        except TypeError:
            b = ntl_GF2E(other, self.c)

        return (op == Py_EQ) == (self.x == b.x)

    def IsZero(ntl_GF2E self):
        """
        Return ``True`` if this element equals zero, ``False`` otherwise.

        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: x = ntl.GF2E([1,0,1,0,1], ctx) ; y = ntl.GF2E([1,1,0,1,1,0,0,0,1], ctx)
            sage: x.IsZero()
            False
            sage: y.IsZero()
            True
        """
        return bool(GF2E_IsZero(self.x))

    def IsOne(ntl_GF2E self):
        """
        Return ``True`` if this element equals one, ``False`` otherwise.

        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: x = ntl.GF2E([1,0,1,0,1], ctx) ; y = ntl.GF2E([0,1,0,1,1,0,0,0,1], ctx)
            sage: x.IsOne()
            False
            sage: y.IsOne()
            True
        """
        return bool(GF2E_IsOne(self.x))

    def trace(ntl_GF2E self):
        """
        Return the trace of this element.

        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: x = ntl.GF2E([1,0,1,0,1], ctx) ; y = ntl.GF2E([0,1,1,0,1,1], ctx)
            sage: x.trace()
            0
            sage: y.trace()
            1
        """
        cdef ntl_GF2 x = ntl_GF2.__new__(ntl_GF2)
        x.x = GF2E_trace(self.x)
        return x

    def rep(ntl_GF2E self):
        """
        Return a ntl.GF2X copy of this element.

        EXAMPLES::

            sage: ctx = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
            sage: a = ntl.GF2E('0x1c', ctx)
            sage: a.rep()
            [1 0 0 0 0 0 1 1]
            sage: type(a.rep())
            <class 'sage.libs.ntl.ntl_GF2X.ntl_GF2X'>
        """
        cdef ntl_GF2X x = ntl_GF2X.__new__(ntl_GF2X)
        x.x = GF2E_rep(self.x)
        return x

    def list(ntl_GF2E self):
        """
        Represent this element as a list of binary digits.

        EXAMPLES::

             sage: e=ntl.GF2E([0,1,1],GF(2^4,'a'))
             sage: e.list()
             [0, 1, 1]
             sage: e=ntl.GF2E('0xff',GF(2^8,'a'))
             sage: e.list()
             [1, 1, 1, 1, 1, 1, 1, 1]

        OUTPUT: list of digits representing the coefficients in this element's
        polynomial representation
        """
        cdef int i
        cdef GF2X_c x = GF2E_rep(self.x)
        cdef ntl_GF2 b

        l = []

        for i from 0 <= i <= GF2X_deg(x):
            b = ntl_GF2.__new__(ntl_GF2)
            b.x = GF2X_coeff(x,i)
            l.append(b)
        return l

    def _sage_(ntl_GF2E self, k=None):
        r"""
        Return a :class:`FiniteFieldElement` representation of this element.

        If a :class:`FiniteField` `k` is provided, it is constructed
        in this field if possible. A :class:`FiniteField` will be
        constructed if none is provided.

        INPUT:

        - ``k`` -- (optional) a field `\GF{2^d}`

        OUTPUT: :class:`FiniteFieldElement` over `k`

        EXAMPLES::

            sage: ctx = ntl.GF2EContext([1,1,0,1,1,0,0,0,1])
            sage: e = ntl.GF2E([0,1], ctx)
            sage: a = e._sage_(); a
            a
        """
        cdef int i
        cdef int length

        self.c.restore_c()

        e = GF2E_degree()

        if k is None:
            from sage.rings.finite_rings.finite_field_constructor import FiniteField
            f = self.c.m._sage_()
            k = FiniteField(2**e, name='a', modulus=f)

        a=k.gen()
        l = self.list()

        length = len(l)
        ret = 0

        for i from 0 <= i < length:
            if l[i]==1:
                ret = ret + a**i

        return ret



# distutils: libraries = flint
# distutils: depends = flint/arb_fmpz_poly.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void _arb_fmpz_poly_evaluate_arb_horner(arb_t res, const fmpz * poly, slong len, const arb_t x, slong prec) noexcept
    void arb_fmpz_poly_evaluate_arb_horner(arb_t res, const fmpz_poly_t poly, const arb_t x, slong prec) noexcept
    void _arb_fmpz_poly_evaluate_arb_rectangular(arb_t res, const fmpz * poly, slong len, const arb_t x, slong prec) noexcept
    void arb_fmpz_poly_evaluate_arb_rectangular(arb_t res, const fmpz_poly_t poly, const arb_t x, slong prec) noexcept
    void _arb_fmpz_poly_evaluate_arb(arb_t res, const fmpz * poly, slong len, const arb_t x, slong prec) noexcept
    void arb_fmpz_poly_evaluate_arb(arb_t res, const fmpz_poly_t poly, const arb_t x, slong prec) noexcept
    void _arb_fmpz_poly_evaluate_acb_horner(acb_t res, const fmpz * poly, slong len, const acb_t x, slong prec) noexcept
    void arb_fmpz_poly_evaluate_acb_horner(acb_t res, const fmpz_poly_t poly, const acb_t x, slong prec) noexcept
    void _arb_fmpz_poly_evaluate_acb_rectangular(acb_t res, const fmpz * poly, slong len, const acb_t x, slong prec) noexcept
    void arb_fmpz_poly_evaluate_acb_rectangular(acb_t res, const fmpz_poly_t poly, const acb_t x, slong prec) noexcept
    void _arb_fmpz_poly_evaluate_acb(acb_t res, const fmpz * poly, slong len, const acb_t x, slong prec) noexcept
    void arb_fmpz_poly_evaluate_acb(acb_t res, const fmpz_poly_t poly, const acb_t x, slong prec) noexcept
    ulong arb_fmpz_poly_deflation(const fmpz_poly_t poly) noexcept
    void arb_fmpz_poly_deflate(fmpz_poly_t res, const fmpz_poly_t poly, ulong deflation) noexcept
    void arb_fmpz_poly_complex_roots(acb_ptr roots, const fmpz_poly_t poly, int flags, slong prec) noexcept
    void arb_fmpz_poly_cos_minpoly(fmpz_poly_t res, ulong n) noexcept
    void arb_fmpz_poly_gauss_period_minpoly(fmpz_poly_t res, ulong q, ulong n) noexcept



# sage.doctest: needs sage.combinat sage.modules
"""
Classical symmetric functions
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>
#                     2012 Mike Zabrocki <mike.zabrocki@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.combinat.partition import _Partitions
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ

from . import hall_littlewood, jack, llt, macdonald, orthotriang, sfa

translate = {'monomial': 'MONOMIAL',
             'homogeneous': 'HOMSYM',
             'powersum': 'POWSYM',
             'elementary': 'ELMSYM',
             'Schur': 'SCHUR'}

conversion_functions = {}


def init():
    """
    Set up the conversion functions between the classical bases.

    EXAMPLES::

        sage: from sage.combinat.sf.classical import init
        sage: sage.combinat.sf.classical.conversion_functions = {}
        sage: init()
        sage: sage.combinat.sf.classical.conversion_functions[('Schur', 'powersum')]
        <cyfunction t_SCHUR_POWSYM_symmetrica at ...>

    The following checks if the bug described in :issue:`15312` is fixed. ::

        sage: change = sage.combinat.sf.classical.conversion_functions[('powersum', 'Schur')]
        sage: hideme = change({Partition([1]*47):ZZ(1)}) # long time
        sage: change({Partition([2,2]):QQ(1)})
        s[1, 1, 1, 1] - s[2, 1, 1] + 2*s[2, 2] - s[3, 1] + s[4]
    """
    import sage.libs.symmetrica.all as symmetrica
    for other_basis, other_name in translate.items():
        for basis, name in translate.items():
            try:
                conversion_functions[(other_basis, basis)] = getattr(symmetrica,
                        f't_{other_name}_{name}')
            except AttributeError:
                pass


init()


###################################
#                                 #
#  Classical Symmetric Functions  #
#                                 #
###################################
class SymmetricFunctionAlgebra_classical(sfa.SymmetricFunctionAlgebra_generic):
    """
    The class of classical symmetric functions.

    .. TODO:: delete this class once all coercions will be handled by Sage's coercion model

    TESTS::

        sage: TestSuite(SymmetricFunctions(QQ).s()).run()
        sage: TestSuite(SymmetricFunctions(QQ).h()).run()
        sage: TestSuite(SymmetricFunctions(QQ).m()).run()
        sage: TestSuite(SymmetricFunctions(QQ).e()).run()
        sage: TestSuite(SymmetricFunctions(QQ).p()).run()
    """

    def _element_constructor_(self, x):
        """
        Convert ``x`` into ``self``, if coercion failed.

        INPUT:

        - ``x`` -- an element of the symmetric functions

        EXAMPLES::

            sage: s = SymmetricFunctions(QQ).s()
            sage: s(2)
            2*s[]
            sage: s([2,1]) # indirect doctest
            s[2, 1]

            sage: McdJ = SymmetricFunctions(QQ['q','t'].fraction_field()).macdonald().J()
            sage: s = SymmetricFunctions(McdJ.base_ring()).s()
            sage: s._element_constructor_(McdJ(s[2,1]))
            s[2, 1]

        TESTS:

        Check that non-Schur bases raise an error when given skew partitions
        (:issue:`19218`)::

            sage: e = SymmetricFunctions(QQ).e()
            sage: e([[2,1],[1]])
            Traceback (most recent call last):
            ...
            TypeError: do not know how to make x (= [[2, 1], [1]]) an element of self

        Check that :issue:`34576` is fixed::

            sage: s = SymmetricFunctions(ZZ).s()
            sage: f = s(0/2); f
            0
            sage: f == 0
            True
            sage: f._monomial_coefficients
            {}

            sage: s2 = SymmetricFunctions(GF(2)).s()
            sage: f = s2(2*s[2,1]); f
            0
            sage: f == 0
            True
            sage: f._monomial_coefficients
            {}
        """
        R = self.base_ring()

        eclass = self.element_class
        if isinstance(x, int):
            x = Integer(x)

        ##############
        # Partitions #
        ##############
        if x in _Partitions:
            return eclass(self, {_Partitions(x): R.one()})

        # Todo: discard all of this which is taken care by Sage's coercion
        # (up to changes of base ring)

        ##############
        # Dual bases #
        ##############
        if isinstance(x, sfa.SymmetricFunctionAlgebra_generic.Element) and hasattr(x, 'dual'):
            # Check to see if it is the dual of some other basis
            # If it is, try to coerce its corresponding element
            # in the other basis
            return self(x.dual())

        ##################################################################
        # Symmetric Functions, same basis, possibly different coeff ring #
        ##################################################################

        # self.Element is used below to test if another symmetric
        # function is expressed in the same basis but in another
        # ground ring.  This idiom is fragile and depends on the
        # internal (unstable) specifications of parents and categories
        #
        # TODO: find the right idiom
        #
        # One cannot use anymore self.element_class: it is build by
        # the category mechanism, and depends on the coeff ring.

        if isinstance(x, self.Element):
            P = x.parent()
            # same base ring
            if P is self:
                return x
            # different base ring
            return eclass(self, {la: rc for la, c in x._monomial_coefficients.items()
                                 if (rc := R(c))})

        ##################################################
        # Classical Symmetric Functions, different basis #
        ##################################################
        if isinstance(x, SymmetricFunctionAlgebra_classical.Element):

            P = x.parent()
            m = x.monomial_coefficients()

            # determine the conversion function.
            try:
                t = conversion_functions[(P.basis_name(), self.basis_name())]
            except AttributeError:
                raise TypeError("do not know how to convert from %s to %s"
                                % (P.basis_name(), self.basis_name()))

            if R == QQ and P.base_ring() == QQ:
                if m:
                    return self._from_dict(t(m)._monomial_coefficients,
                                           coerce=True)
                return self.zero()
            f = lambda part: self._from_dict(t({part: ZZ.one()})._monomial_coefficients)
            return self._apply_module_endomorphism(x, f)

        ###############################
        # Hall-Littlewood Polynomials #
        ###############################
        if isinstance(x, hall_littlewood.HallLittlewood_generic.Element):
            #
            # Qp: Convert to Schur basis and then convert to self
            #
            if isinstance(x, (hall_littlewood.HallLittlewood_qp.Element,
                              hall_littlewood.HallLittlewood_p.Element)):
                P = x.parent()
                sx = P._s._from_cache(x, P._s_cache, P._self_to_s_cache, t=P.t)
                return self(sx)
            #
            # Q: Convert to P basis and then convert to self
            #
            if isinstance(x, hall_littlewood.HallLittlewood_q.Element):
                return self(x.parent()._P(x))

        #######
        # LLT #
        #######
        # Convert to m and then to self.
        elif isinstance(x, llt.LLT_generic.Element):
            P = x.parent()
            Rx = P.base_ring()
            zero = R.zero()
            if not R.has_coerce_map_from(Rx):
                raise TypeError("no coerce map from x's parent's base ring (= %s) to self's base ring (= %s)"
                                % (Rx, R))

            z_elt = {}
            for m, c in x._monomial_coefficients.items():
                n = sum(m)
                P._m_cache(n)
                for part in P._self_to_m_cache[n][m]:
                    z_elt[part] = z_elt.get(part, zero) + R(c*P._self_to_m_cache[n][m][part].subs(t=P.t))

            m = P._sym.monomial()
            return self(m._from_dict(z_elt))

        #########################
        # Macdonald Polynomials #
        #########################
        elif isinstance(x, macdonald.MacdonaldPolynomials_generic.Element):
            if isinstance(x, (macdonald.MacdonaldPolynomials_j.Element,
                              macdonald.MacdonaldPolynomials_s.Element)):
                P = x.parent()
                sx = P._s._from_cache(x, P._s_cache, P._self_to_s_cache, q=P.q, t=P.t)
                return self(sx)
            if isinstance(x, (macdonald.MacdonaldPolynomials_q.Element,
                                macdonald.MacdonaldPolynomials_p.Element)):
                J = x.parent()._J
                jx = J(x)
                sx = J._s._from_cache(jx, J._s_cache, J._self_to_s_cache, q=J.q, t=J.t)
                return self(sx)
            if isinstance(x, (macdonald.MacdonaldPolynomials_h.Element,
                                macdonald.MacdonaldPolynomials_ht.Element)):
                P = x.parent()
                sx = P._self_to_s(x)
                return self(sx)
            raise TypeError

        ####################
        # Jack Polynomials #
        ####################
        elif isinstance(x, jack.JackPolynomials_generic.Element):
            if isinstance(x, jack.JackPolynomials_p.Element):
                P = x.parent()
                mx = P._m._from_cache(x, P._m_cache, P._self_to_m_cache, t=P.t)
                return self(mx)
            if isinstance(x, (jack.JackPolynomials_j.Element,
                              jack.JackPolynomials_q.Element)):
                return self(x.parent()._P(x))
            raise TypeError

        ####################################################
        # Bases defined by orthogonality and triangularity #
        ####################################################
        elif isinstance(x, orthotriang.SymmetricFunctionAlgebra_orthotriang.Element):
            # Convert to its base and then to self
            P = x.parent()
            if self is P._sf_base:
                return P._sf_base._from_cache(x, P._base_cache, P._self_to_base_cache)
            return self( P._sf_base(x) )

        #################################
        # Last shot -- try calling R(x) #
        #################################
        else:
            try:
                c = R(x)
            except (TypeError, ValueError):
                raise TypeError("do not know how to make x (= {}) an element of self".format(x))
            else:
                if not c:
                    return self.zero()
                return eclass(self, {_Partitions([]): c})

    # This subclass is currently needed for the test above:
    #    isinstance(x, SymmetricFunctionAlgebra_classical.Element):
    class Element(sfa.SymmetricFunctionAlgebra_generic.Element):
        """
        A symmetric function.
        """
        pass



from sage.categories.map cimport Map
from sage.structure.parent cimport Parent


cdef class RingExtension_generic(Parent):
    cdef _type
    cdef _backend
    cdef _defining_morphism
    cdef _backend_defining_morphism
    cdef dict _print_options
    cdef bint _import_methods
    cdef bint _is_backend_exposed
    # For division
    cdef RingExtension_generic _fraction_field
    cdef type _fraction_field_type

    cpdef is_defined_over(self, base)
    cpdef Parent _check_base(self, Parent base)
    cpdef _degree_over(self, Parent base)
    cpdef _is_finite_over(self, Parent base)
    cpdef _is_free_over(self, Parent base)
    cdef Map _defining_morphism_fraction_field(self, bint extend_base)


cdef class RingExtensionFractionField(RingExtension_generic):
    cdef _ring


cdef class RingExtensionWithBasis(RingExtension_generic):
    cdef _basis
    cdef _basis_names
    cdef _basis_latex_names

    cpdef _basis_over(self, Parent base)
    # cpdef _free_module(self, Parent base, bint map)


cdef class RingExtensionWithGen(RingExtensionWithBasis):
    cdef _gen
    cdef _name
    cdef public object _latex_names



r"""
Cubic Hecke matrix representations

This module contains the class :class:`CubicHeckeMatrixRep` which is used to
treat the matrix representations of the elements of the cubic Hecke algebra
(:class:`~sage.algebras.hecke_algebras.cubic_hecke_algebra.CubicHeckeAlgebra`)
together with its parent class :class:`CubicHeckeMatrixSpace`. Furthermore,
it contains enums for their types (:class:`RepresentationType`) and names
(:class:`AbsIrreducibeRep`).


AUTHORS:

- Sebastian Oehms May 2020: initial version
"""
# ###########################################################################
#       Copyright (C) 2020 Sebastian Oehms <seb.oehms@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ###########################################################################
from enum import Enum

from sage.misc.cachefunc import cached_method
from sage.misc.verbose import verbose
from sage.rings.integer import Integer
from sage.matrix.matrix_generic_dense import Matrix_generic_dense
from sage.matrix.matrix_space import MatrixSpace
from sage.matrix.constructor import matrix
from sage.matrix.special import block_diagonal_matrix
from sage.databases.cubic_hecke_db import CubicHeckeDataSection as sc


# -------------------------------------------
# Enum for the generators sign (of exponent)
# -------------------------------------------
class GenSign(Enum):
    r"""
    Enum class to select the braid generators sign.

    EXAMPLES::

        sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
        sage: chmr.GenSign.pos
        <GenSign.pos: 1>
        sage: chmr.GenSign.neg
        <GenSign.neg: -1>
    """
    pos = 1
    neg = -1


# -------------------------------------------
# Enum for typ of matrix representation
# -------------------------------------------
class RepresentationType(Enum):
    r"""
    Enum class to select a representation type for the cubic Hecke algebra.

    - ``RegularLeft`` -- left regular representations
    - ``RegularRight`` -- right regular representations
    - ``SplitIrredMarin`` -- split irreducible representations obtained from
      Ivan Marin's data
    - ``SplitIrredChevie`` -- the split irreducible representations obtained
      from CHEVIE via the ``GAP3`` interface

    EXAMPLES::

        sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
        sage: chmr.RepresentationType.RegularLeft.is_regular()
        True
    """
    def is_split(self) -> bool:
        r"""
        Return ``True`` if this representation type is absolutely split,
        ``False`` else-wise.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: chevie = chmr.RepresentationType.SplitIrredChevie
            sage: chevie.is_split()
            True
        """
        return self.value['split']

    def is_regular(self) -> bool:
        r"""
        Return ``True`` if this representation type is regular, ``False``
        else-wise.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: reg_left = chmr.RepresentationType.RegularLeft
            sage: reg_left.is_regular()
            True
        """
        return self.value['regular']

    def data_section(self):
        r"""
        Return the name of the data file. For more information see
        :class:`~sage.databases.cubic_hecke_db.CubicHeckeDataBase`.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: reg_left = chmr.RepresentationType.RegularLeft
            sage: reg_left.data_section()
            <CubicHeckeDataSection.regular_left: 'regular_left'>
        """
        return self.value['data']

    def number_of_representations(self, nstrands):
        r"""
        Return the number of representations existing to that type.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: chmr.RepresentationType.SplitIrredChevie.number_of_representations(4)
            24
            sage: chmr.RepresentationType.SplitIrredMarin.number_of_representations(4)
            24
        """
        if self.value['data'] is None:
            if nstrands < 1 or nstrands > 5:
                raise ValueError("nstrands must be between 1 and 5")
        elif nstrands < 1 or nstrands > 4:
            raise ValueError("nstrands must be between 1 and 4")
        return self.value['num_rep'][nstrands - 1]

    RegularLeft = {'split': False, 'regular': True, 'data': sc.regular_left, 'num_rep': [1, 1, 1, 1]}
    RegularRight = {'split': False, 'regular': True, 'data': sc.regular_right, 'num_rep': [1, 1, 1, 1]}
    SplitIrredMarin = {'split': True, 'regular': False, 'data': sc.split_irred, 'num_rep': [1, 3, 7, 24]}
    SplitIrredChevie = {'split': True, 'regular': False, 'data': None, 'num_rep': [1, 3, 7, 24, 30]}


# ---------------------------------------------
# Enum for absolute irreducible representations
# ---------------------------------------------
class AbsIrreducibeRep(Enum):
    r"""
    Enum class to select an absolutely irreducible representation for the cubic
    Hecke algebra (``CHAn``) on `n`-strands.

    The names are build as follows: Take the determinant of one of the
    generators of the ``CHAn``. This is a monomial in the generic extension
    ring (``GER``) of ``CHA``, say ``a^ib^jc^k`` where ``a, b`` and ``c`` are
    the generators of ``GER``. This does not depend on the choice of the
    generator of ``CHA``, since these are conjugated to each other. This
    monomial might be looked as the weight of the representation. Therefore we
    use it as a name:

        ``Wn_ijk``

    The only ambiguity among the available irreducible representations occurs for the two nine-dimensional modules, which
    are conjugated to each other and distinguished by these names:

        ``W4_333`` and ``W4_333bar``

    Examples of names:

    - ``W2_100`` -- one dimensional representation of the cubic Hecke algebra on 2 strands corresponding to the first root
      of the cubic equation
    - ``W3_111`` -- three dimensional irreducible representation of the cubic Hecke algebra on 3 strands
    - ``W4_242`` -- eight dimensional irreducible representation of the cubic Hecke algebra on 4 strands having the second
      root of the cubic equation as weight of dimension 4

    Alternative names are taken from [MW2012]_ and can be shown by
    :meth:`alternative_name`.

    EXAMPLES::

        sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
        sage: [irr.name for irr in chmr.AbsIrreducibeRep]
        ['W2_100', 'W2_001', 'W2_010', 'W3_100', 'W3_001', 'W3_010', 'W3_011', 'W3_110',
         'W3_101', 'W3_111', 'W4_100', 'W4_001', 'W4_010', 'W4_011', 'W4_110', 'W4_101',
         'W4_111', 'W4_120', 'W4_201', 'W4_012', 'W4_102', 'W4_210', 'W4_021', 'W4_213',
         'W4_132', 'W4_321', 'W4_231', 'W4_123', 'W4_312', 'W4_422', 'W4_224', 'W4_242',
         'W4_333', 'W4_333bar', 'W5_100', 'W5_001', 'W5_010', 'W5_013', 'W5_130', 'W5_301',
         'W5_031', 'W5_103', 'W5_310', 'W5_203', 'W5_032', 'W5_320', 'W5_230', 'W5_023',
         'W5_302', 'W5_033', 'W5_330', 'W5_303', 'W5_163', 'W5_631', 'W5_316', 'W5_136',
         'W5_613', 'W5_361', 'W5_366', 'W5_663', 'W5_636', 'W5_933', 'W5_339', 'W5_393']

    REFERENCES:

    - [MW2012]_
    """
    def alternative_name(self):
        r"""
        Return the name of the split irreducible representation for cubic Hecke
        algebras for up to four strands as given in [MW2012]_.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: chmr.AbsIrreducibeRep.W3_011.alternative_name()
            'Tbc'
        """
        return self.value['alt_name']

    def dimension(self):
        r"""
        Return the dimension of the representation.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: chmr.AbsIrreducibeRep.W3_111.dimension()
            3
        """
        return self.value['dim']

    def number_gens(self):
        r"""
        Return the number of generators of the underlying cubic Hecke algebra.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: chmr.AbsIrreducibeRep.W3_001.number_gens()
            2
            sage: chmr.AbsIrreducibeRep.W4_001.number_gens()
            3
        """
        return self.value['ngens']

    def length_orbit(self):
        r"""
        Return the length of the orbit of this representation under the action
        of the Galois group of the cubic equation.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: chmr.AbsIrreducibeRep.W3_001.length_orbit()
            3
            sage: chmr.AbsIrreducibeRep.W3_111.length_orbit()
            1
        """
        return self.value['len_orbit']

    def gap_index(self):
        r"""
        Return the array index of this representation for the access
        to the ``GAP3`` package ``CHEVIE``.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: chmr.AbsIrreducibeRep.W3_111.gap_index()
            6
        """
        return self.value['gap_ind']

    def internal_index(self):
        r"""
        Return the array index of this representation for the internal access.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: chmr.AbsIrreducibeRep.W3_111.internal_index()
            6
        """
        return self.value['intern_ind']

    # -------------------------------------------------------------------------------------------------
    # absolutely irreducible representations corresponding to braids on 2 strands
    # -------------------------------------------------------------------------------------------------
    W2_100 = {'alt_name': 'Sa', 'dim': 1, 'ngens': 1, 'len_orbit': 3, 'gap_ind': 0, 'intern_ind': 0}
    W2_001 = {'alt_name': 'Sc', 'dim': 1, 'ngens': 1, 'len_orbit': 3, 'gap_ind': 1, 'intern_ind': 1}
    W2_010 = {'alt_name': 'Sb', 'dim': 1, 'ngens': 1, 'len_orbit': 3, 'gap_ind': 2, 'intern_ind': 2}

    # -------------------------------------------------------------------------------------------------
    # absolutely irreducible representations corresponding to braids on 3 strands
    # -------------------------------------------------------------------------------------------------
    W3_100 = {'alt_name': 'Sa', 'dim': 1, 'ngens': 2, 'len_orbit': 3, 'gap_ind': 0, 'intern_ind': 0}
    W3_001 = {'alt_name': 'Sc', 'dim': 1, 'ngens': 2, 'len_orbit': 3, 'gap_ind': 1, 'intern_ind': 1}
    W3_010 = {'alt_name': 'Sb', 'dim': 1, 'ngens': 2, 'len_orbit': 3, 'gap_ind': 2, 'intern_ind': 2}

    W3_011 = {'alt_name': 'Tbc', 'dim': 2, 'ngens': 2, 'len_orbit': 3, 'gap_ind': 3, 'intern_ind': 3}
    W3_110 = {'alt_name': 'Tab', 'dim': 2, 'ngens': 2, 'len_orbit': 3, 'gap_ind': 4, 'intern_ind': 4}
    W3_101 = {'alt_name': 'Tac', 'dim': 2, 'ngens': 2, 'len_orbit': 3, 'gap_ind': 5, 'intern_ind': 5}

    W3_111 = {'alt_name': 'V', 'dim': 3, 'ngens': 2, 'len_orbit': 1, 'gap_ind': 6, 'intern_ind': 6}

    # -------------------------------------------------------------------------------------------------
    # absolutely irreducible representations corresponding to braids on 4 strands
    # -------------------------------------------------------------------------------------------------
    W4_100 = {'alt_name': 'Sa', 'dim': 1, 'ngens': 3, 'len_orbit': 3, 'gap_ind': 0, 'intern_ind': 0}
    W4_001 = {'alt_name': 'Sc', 'dim': 1, 'ngens': 3, 'len_orbit': 3, 'gap_ind': 1, 'intern_ind': 1}
    W4_010 = {'alt_name': 'Sb', 'dim': 1, 'ngens': 3, 'len_orbit': 3, 'gap_ind': 2, 'intern_ind': 2}

    W4_011 = {'alt_name': 'Tbc', 'dim': 2, 'ngens': 3, 'len_orbit': 3, 'gap_ind': 3, 'intern_ind': 3}
    W4_110 = {'alt_name': 'Tab', 'dim': 2, 'ngens': 3, 'len_orbit': 3, 'gap_ind': 4, 'intern_ind': 4}
    W4_101 = {'alt_name': 'Tac', 'dim': 2, 'ngens': 3, 'len_orbit': 3, 'gap_ind': 5, 'intern_ind': 5}

    W4_111 = {'alt_name': 'V', 'dim': 3, 'ngens': 3, 'len_orbit': 1, 'gap_ind': 6, 'intern_ind': 6}

    W4_120 = {'alt_name': 'Uba', 'dim': 3, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 7, 'intern_ind': 7}
    W4_201 = {'alt_name': 'Uac', 'dim': 3, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 8, 'intern_ind': 8}
    W4_012 = {'alt_name': 'Ucb', 'dim': 3, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 9, 'intern_ind': 9}
    W4_102 = {'alt_name': 'Uca', 'dim': 3, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 10, 'intern_ind': 10}
    W4_210 = {'alt_name': 'Uab', 'dim': 3, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 11, 'intern_ind': 11}
    W4_021 = {'alt_name': 'Ubc', 'dim': 3, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 12, 'intern_ind': 12}

    W4_213 = {'alt_name': 'Vcab', 'dim': 6, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 13, 'intern_ind': 13}
    W4_132 = {'alt_name': 'Vbca', 'dim': 6, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 14, 'intern_ind': 14}
    W4_321 = {'alt_name': 'Vabc', 'dim': 6, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 15, 'intern_ind': 15}
    W4_231 = {'alt_name': 'Vbac', 'dim': 6, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 16, 'intern_ind': 16}
    W4_123 = {'alt_name': 'Vcba', 'dim': 6, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 17, 'intern_ind': 17}
    W4_312 = {'alt_name': 'Vacb', 'dim': 6, 'ngens': 3, 'len_orbit': 6, 'gap_ind': 18, 'intern_ind': 18}

    W4_422 = {'alt_name': 'Wa', 'dim': 8, 'ngens': 3, 'len_orbit': 3, 'gap_ind': 19, 'intern_ind': 19}
    W4_224 = {'alt_name': 'Wc', 'dim': 8, 'ngens': 3, 'len_orbit': 3, 'gap_ind': 20, 'intern_ind': 20}
    W4_242 = {'alt_name': 'Wb', 'dim': 8, 'ngens': 3, 'len_orbit': 3, 'gap_ind': 21, 'intern_ind': 21}

    W4_333 = {'alt_name': 'X', 'dim': 9, 'ngens': 3, 'len_orbit': 2, 'gap_ind': 22, 'intern_ind': 22}
    W4_333bar = {'alt_name': 'Xbar', 'dim': 9, 'ngens': 3, 'len_orbit': 2, 'gap_ind': 23, 'intern_ind': 23}

    # -------------------------------------------------------------------------------------------------
    # absolutely irreducible representations corresponding to braids on 5 strands
    # -------------------------------------------------------------------------------------------------
    W5_100 = {'alt_name': None, 'dim': 1, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 0, 'intern_ind': 0}
    W5_001 = {'alt_name': None, 'dim': 1, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 1, 'intern_ind': 1}
    W5_010 = {'alt_name': None, 'dim': 1, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 2, 'intern_ind': 2}

    W5_013 = {'alt_name': None, 'dim': 4, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 3, 'intern_ind': 3}
    W5_130 = {'alt_name': None, 'dim': 4, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 4, 'intern_ind': 4}
    W5_301 = {'alt_name': None, 'dim': 4, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 5, 'intern_ind': 5}
    W5_031 = {'alt_name': None, 'dim': 4, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 6, 'intern_ind': 6}
    W5_103 = {'alt_name': None, 'dim': 4, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 7, 'intern_ind': 7}
    W5_310 = {'alt_name': None, 'dim': 4, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 8, 'intern_ind': 8}

    W5_203 = {'alt_name': None, 'dim': 5, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 9, 'intern_ind': 9}
    W5_032 = {'alt_name': None, 'dim': 5, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 10, 'intern_ind': 10}
    W5_320 = {'alt_name': None, 'dim': 5, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 11, 'intern_ind': 11}
    W5_230 = {'alt_name': None, 'dim': 5, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 12, 'intern_ind': 12}
    W5_023 = {'alt_name': None, 'dim': 5, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 13, 'intern_ind': 13}
    W5_302 = {'alt_name': None, 'dim': 5, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 14, 'intern_ind': 14}

    W5_033 = {'alt_name': None, 'dim': 6, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 15, 'intern_ind': 15}
    W5_330 = {'alt_name': None, 'dim': 6, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 16, 'intern_ind': 16}
    W5_303 = {'alt_name': None, 'dim': 6, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 17, 'intern_ind': 17}

    W5_163 = {'alt_name': None, 'dim': 10, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 18, 'intern_ind': 18}
    W5_631 = {'alt_name': None, 'dim': 10, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 19, 'intern_ind': 19}
    W5_316 = {'alt_name': None, 'dim': 10, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 20, 'intern_ind': 20}
    W5_136 = {'alt_name': None, 'dim': 10, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 21, 'intern_ind': 21}
    W5_613 = {'alt_name': None, 'dim': 10, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 22, 'intern_ind': 22}
    W5_361 = {'alt_name': None, 'dim': 10, 'ngens': 4, 'len_orbit': 6, 'gap_ind': 23, 'intern_ind': 23}

    W5_366 = {'alt_name': None, 'dim': 15, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 24, 'intern_ind': 24}
    W5_663 = {'alt_name': None, 'dim': 15, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 26, 'intern_ind': 25}
    W5_636 = {'alt_name': None, 'dim': 15, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 27, 'intern_ind': 26}

    W5_933 = {'alt_name': None, 'dim': 15, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 25, 'intern_ind': 27}
    W5_339 = {'alt_name': None, 'dim': 15, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 28, 'intern_ind': 28}
    W5_393 = {'alt_name': None, 'dim': 15, 'ngens': 4, 'len_orbit': 3, 'gap_ind': 29, 'intern_ind': 29}


# ------------------------------------------------------------------------------------------------------------------
# Definition of CubicHeckeMatrixRep
# --------------------------------------------------------------------------------------------------------
class CubicHeckeMatrixRep(Matrix_generic_dense):
    r"""
    Class to supervise the diagonal block matrix structure arising from
    cubic Hecke algebra-representations.

    EXAMPLES::

        sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
        sage: CHA2.<c1> = algebras.CubicHecke(2)
        sage: MS = chmr.CubicHeckeMatrixSpace(CHA2)
        sage: m1 = MS(c1); m1
        [         a          0          0]
        [         0          b          0]
        [         0          0 -b - a + u]
        sage: type(m1)
        <class 'sage.algebras.hecke_algebras.cubic_hecke_matrix_rep.CubicHeckeMatrixSpace_with_category.element_class'>
        sage: m1.block_diagonal_list()
        [[a], [b], [-b - a + u]]

        sage: MSo = chmr.CubicHeckeMatrixSpace(CHA2, original=True)
        sage: MSo(c1)
        [a 0 0]
        [0 b 0]
        [0 0 c]

        sage: reg_left = chmr.RepresentationType.RegularLeft
        sage: MSreg = chmr.CubicHeckeMatrixSpace(CHA2, representation_type=reg_left)
        sage: MSreg(c1)
        [ 0 -v  1]
        [ 1  u  0]
        [ 0  w  0]
        sage: len(_.block_diagonal_list())
        1

    TESTS:

    The minpoly does not work over more generic rings::

        sage: TestSuite(m1).run(skip='_test_minpoly')
    """

    @cached_method
    def _get_block(self, ind):
        r"""
        Return the ``ind``-th sub-matrix block of ``self`` considered
        as block diagonal matrix.

        INPUT:

        - ``ind`` -- integer specifying the list index according to
          :meth:`internal_index` respectively :meth:`gap_index`

        OUTPUT:

        An instance of :class:`Matrix_generic_dense` representing
        the specified block of ``self``.

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: c1.matrix()._get_block(0)      # indirect doctest
            [a]
        """
        representation_type = self.parent()._representation_type
        if not representation_type.is_split():
            return matrix(self)
        n = self.parent()._cubic_hecke_algebra.ngens()
        s = sum(irr_rep.dimension() for irr_rep in AbsIrreducibeRep if irr_rep.number_gens() == n and irr_rep.internal_index() < ind)
        for irr_rep in AbsIrreducibeRep:
            if irr_rep.number_gens() == n and irr_rep.internal_index() == ind:
                d = irr_rep.dimension()
                return matrix(self.submatrix(s, s, d, d))
        raise ValueError('no irreducible representation for this index')

    @cached_method
    def _irr_to_ind(self, irr):
        r"""
        Return the index if the given split irreducible representation
        of ``self``.

        INPUT:

        - ``irr`` -- an instance of :class:`AbsIrreducibeRep` specifying an
          absolute irreducible representation of the cubic Hecke algebra

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: m1 = c1.matrix()
            sage: m1._irr_to_ind(CHA2.irred_repr.W2_001)
            1
            sage: m1._irr_to_ind(CHA2.irred_repr.W3_001)
            Traceback (most recent call last):
            ...
            TypeError: representation must have 1 generators
        """
        representation_type = self.parent()._representation_type
        if not representation_type.is_split():
            raise TypeError('representation type is non split')

        ch_algebra = self.parent()._cubic_hecke_algebra
        if ch_algebra.strands() != irr.number_gens() + 1:
            raise TypeError('representation must have %s generators' % (ch_algebra.strands() - 1))

        ind = irr.gap_index()
        if representation_type == RepresentationType.SplitIrredMarin:
            ind = irr.internal_index()
        return ind

    @cached_method
    def __getitem__(self, item):
        r"""
        Return the sub-matrix block of ``self`` considered as block diagonal
        matrix specified by `item`.

        Overloading builtin-method to select a list-item.

        INPUT:

        - ``item`` -- an :class:`AbsIrreducibeRep` specifying an
          absolute irreducible representation of the cubic Hecke algebra;
          alternatively, it can be specified by list index
          (see :meth:`internal_index` respectively :meth:`gap_index`)

        OUTPUT:

        An instance of :class:`Matrix_generic_dense` representing
        the specified block of ``self``.

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: m1 = c1.matrix()
            sage: m1[0]                       # indirect doctest
            [a]
            sage: m1[CHA2.irred_repr.W2_001]  # indirect doctest
            [b]
        """
        if isinstance(item, AbsIrreducibeRep):
            return self._get_block(self._irr_to_ind(item))
        if isinstance(item, (Integer, int)):
            return self._get_block(item)

        return super().__getitem__(item)

    @cached_method
    def block_diagonal_list(self):
        r"""
        Return the list of sub-matrix blocks of ``self`` considered
        as block diagonal matrix.

        OUTPUT:

        A list of instances of :class:`Matrix_generic_dense` each of
        which represents a diagonal block of ``self``.

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: c1.matrix().block_diagonal_list()
            [[a], [b], [-b - a + u]]
        """
        representation_type = self.parent()._representation_type
        n = self.parent()._cubic_hecke_algebra.strands()
        m = representation_type.number_of_representations(n)
        return [self._get_block(i) for i in range(m)]

    @cached_method
    def reduce_to_irr_block(self, irr):
        r"""
        Return a copy of ``self`` with zeroes outside the block corresponding to
        ``irr`` but the block according to the input identical to that of ``self``.

        INPUT:

        - ``irr`` -- an :class:`AbsIrreducibeRep` specifying an
          absolute irreducible representation of the cubic Hecke algebra;
          alternatively, it can be specified by list index (see
          :meth:`internal_index` respectively :meth:`gap_index`)

        OUTPUT:

        An instance of :class:`Matrix_generic_dense` with exactly one nonzero block
        according to ``irr``.

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: m1 = c1.matrix()
            sage: m1.reduce_to_irr_block(0)
            [a 0 0]
            [0 0 0]
            [0 0 0]
            sage: m1.reduce_to_irr_block(CHA2.irred_repr.W2_001)
            [0 0 0]
            [0 b 0]
            [0 0 0]
        """
        if isinstance(irr, AbsIrreducibeRep):
            ind = self._irr_to_ind(irr)
        else:
            ind = Integer(irr)
        from copy import copy
        mat_list = copy(self.parent().zero().block_diagonal_list())
        mat_list[ind] = self[ind]
        return block_diagonal_matrix(mat_list, subdivide=self.parent()._subdivide, sparse=True)


# ------------------------------------------------------------------------------------------------------------------
# Definition of CubicHeckeMatrixSpace
# --------------------------------------------------------------------------------------------------------
class CubicHeckeMatrixSpace(MatrixSpace):
    r"""
    The matrix space of cubic Hecke algebra representations.

    INPUT:

    - ``cubic_hecke_algebra`` -- (optional)
      :class:`~sage.algebras.hecke_algebras.cubic_hecke_algebra.CubicHeckeAlgebra`
      must be given if ``element`` fails to be an instance of its element class
    - ``representation_type`` -- (default: ``RepresentationType.SplitIrredChevie``)
      :class:`RepresentationType` specifying the type of the representation
    - ``subdivide`` -- boolean (default: ``False``); whether or not to subdivide
      the resulting matrices

    - ``original`` -- boolean (default: ``False``); if ``True``, the matrix
      will have coefficients in the generic base / extension ring

    EXAMPLES::

        sage: CHA2.<c1> = algebras.CubicHecke(2)
        sage: c1.matrix()      # indirect doctest
        [         a          0          0]
        [         0          b          0]
        [         0          0 -b - a + u]
        sage: c1.matrix(original=True)
        [a 0 0]
        [0 b 0]
        [0 0 c]
        sage: c1.matrix(representation_type = CHA2.repr_type.RegularLeft)   # indirect doctest
        [ 0 -v  1]
        [ 1  u  0]
        [ 0  w  0]
    """
    @staticmethod
    def __classcall_private__(cls, cubic_hecke_algebra, representation_type=None, subdivide=False, original=False):
        r"""
        Normalize the arguments to call the ``__init__`` constructor.

        See the documentation in ``__init__``.

        TESTS::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: MS = chmr.CubicHeckeMatrixSpace(CHA2)
            sage: MS2 = chmr.CubicHeckeMatrixSpace(CHA2, representation_type=CHA2.repr_type.SplitIrredMarin, subdivide=False)
            sage: MS is MS2
            True
        """
        from sage.algebras.hecke_algebras.cubic_hecke_algebra import CubicHeckeAlgebra

        if not isinstance(cubic_hecke_algebra, CubicHeckeAlgebra):
            raise TypeError('cubic_hecke_algebra must be an instance of CubicHeckeAlgebra')

        if representation_type is None:
            representation_type = RepresentationType.SplitIrredMarin

        if representation_type == RepresentationType.SplitIrredChevie:
            from sage.combinat.root_system.reflection_group_real import is_chevie_available
            if not is_chevie_available():
                raise ValueError('CHEVIE is not available')

        base_ring = cubic_hecke_algebra.base_ring(generic=original)
        dimension = cubic_hecke_algebra.dimension()
        if representation_type.is_split():
            dimension = cubic_hecke_algebra._dim_irr_rep
            base_ring = cubic_hecke_algebra.extension_ring(generic=original)
        # Bypass the MatrixSpace.__classcall__
        return super(MatrixSpace, cls).__classcall__(cls, base_ring, int(dimension),
                                                     cubic_hecke_algebra=cubic_hecke_algebra,
                                                     representation_type=representation_type,
                                                     subdivide=subdivide)

    def __init__(self, base_ring,
                 dimension,
                 cubic_hecke_algebra,
                 representation_type,
                 subdivide):
        r"""
        Initialize ``self``.

        TESTS::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: CHA3.<c1, c2> = algebras.CubicHecke(3)
            sage: MS = chmr.CubicHeckeMatrixSpace(CHA3, original=True)

        The minpoly does not work over more generic rings::

            sage: TestSuite(MS).run(skip='_test_elements')     # long time
        """
        from sage.algebras.hecke_algebras.cubic_hecke_algebra import CubicHeckeAlgebra

        if not isinstance(cubic_hecke_algebra, CubicHeckeAlgebra):
            raise TypeError('cubic_hecke_algebra must be an instance of CubicHeckeAlgebra')

        # -------------------------------------------------------------------------------------------------
        # saving input parameters
        # -------------------------------------------------------------------------------------------------
        self._cubic_hecke_algebra = cubic_hecke_algebra
        self._representation_type = representation_type
        self._subdivide = subdivide

        original_base_ring = cubic_hecke_algebra.base_ring(generic=True)

        if representation_type.is_split():
            original_base_ring = cubic_hecke_algebra.extension_ring(generic=True)
            specialize = cubic_hecke_algebra._generic_extension_ring_map
        else:
            specialize = cubic_hecke_algebra._ring_of_definition_map

        verbose("original_base_ring %s base_ring %s" % (original_base_ring, base_ring), level=2)

        self._original_base_ring = original_base_ring
        self._specialize = specialize

        super().__init__(base_ring, dimension, dimension, sparse=True, implementation=CubicHeckeMatrixRep)

    def construction(self):
        r"""
        Return ``None`` since this construction is not functorial.

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: MS = c1.matrix().parent()
            sage: MS._test_category()   # indirect doctest
        """
        return

    def __reduce__(self):
        r"""
        Used for pickling.

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: MS = c1.matrix().parent()
            sage: loads(dumps(MS)) == MS      # indirect doctest
            True
        """
        original = self.base_ring() == self._original_base_ring
        return CubicHeckeMatrixSpace, (self._cubic_hecke_algebra, self._representation_type, self._subdivide, original)

    def _element_constructor_(self, x):
        r"""
        INPUT:

        - ``x`` -- an element of a
          :class:`~sage.algebras.hecke_algebras.cubic_hecke_algebra.CubicHeckeAlgebra`
          or an element whose parent is a :class:`MatrixSpace`

        EXAMPlES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: CHA3.<c1, c2> = algebras.CubicHecke(3)
            sage: MS = chmr.CubicHeckeMatrixSpace(CHA3, original=True)
            sage: m1 = MS._element_constructor_(c1)
            sage: isinstance(m1, MS.element_class)
            True
            sage: isinstance(MS._element_constructor_(m1), MS.element_class)
            True

            sage: m = matrix(MS.base_ring(), 12, 12, lambda i, j: 1)
            sage: MS._element_constructor_(m)
            Traceback (most recent call last):
            ...
            TypeError: incompatible block structure
        """
        # -------------------------------------------------------------------------------------------------
        # checking input and setting the self._cubic_hecke_algebra
        # -------------------------------------------------------------------------------------------------
        ch_algebra = self._cubic_hecke_algebra
        ele_parent = x.parent()
        ori_base_ring = self._original_base_ring
        if isinstance(ele_parent, MatrixSpace):
            # TODO: Find preimage in cubic hecke algebra
            d1, d2 = x.dimensions()
            if d1 != self.ncols() or d2 != self.nrows():
                raise ValueError('incompatible dimensions!')

            if ele_parent.base_ring() == ori_base_ring:
                x = self._specialize_matrix(x)
            elif ele_parent.base_ring() != self.base_ring():
                raise ValueError('incompatible base ring!')
            x_in_self = self.element_class(self, x)
            matrix_list = x_in_self.block_diagonal_list()
            matrix = block_diagonal_matrix(matrix_list, subdivide=self._subdivide, sparse=True)
            if matrix != x:
                raise TypeError('incompatible block structure')
            return self.element_class(self, matrix)

        if ele_parent == ch_algebra:
            mat = ch_algebra._apply_module_morphism(x, self._image_on_basis)
            return self(mat)

        raise TypeError('element must be an instance of CubicHeckeElement or a matrix')

    @cached_method
    def __call__(self, entries=None, coerce=True, copy=None):
        r"""
        Construct an element of ``self``.

        This method needs to be overloaded here since
        :class:`MatrixSpace` has an own implementation of it.

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: MS = chmr.CubicHeckeMatrixSpace(CHA2)
            sage: MS(c1)
            [         a          0          0]
            [         0          b          0]
            [         0          0 -b - a + u]
        """
        from sage.algebras.hecke_algebras.cubic_hecke_algebra import CubicHeckeAlgebra
        if entries is None:
            return super().__call__(entries=entries, coerce=coerce, copy=copy)
        if not hasattr(entries, 'parent'):
            return super().__call__(entries=entries, coerce=coerce, copy=copy)
        ele_parent = entries.parent()
        if not isinstance(ele_parent, (CubicHeckeAlgebra, MatrixSpace)):
            return super().__call__(entries=entries, coerce=coerce, copy=copy)
        return self._element_constructor_(entries)

    @cached_method
    def _specialize_matrix(self, mat):
        r"""
        Return the given matrix specializing the original coefficients
        from data import to the base ring of ``self``.

        INPUT:

        - ``mat`` -- matrix over the original base ring

        OUTPUT: matrix over the base ring of ``self``

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: MS = chmr.CubicHeckeMatrixSpace(CHA2)
            sage: B = MS._original_base_ring
            sage: a, b, c = B.gens()
            sage: mat = matrix(B, [[a, b], [0, c]])
            sage: MS._specialize_matrix(mat)
            [         a          b]
            [         0 -b - a + u]
        """
        base_ring = self.base_ring()
        original_base_ring = self._original_base_ring
        specialize = self._specialize

        if base_ring == original_base_ring:
            return mat

        mat_dict = {k: specialize(original_base_ring(v)) for k, v in mat.dict().items()}
        return matrix(base_ring, mat_dict)

    @cached_method
    def _image_on_gen(self, gen_ind):
        r"""
        Return the matrix list corresponding to the generator given by
        ``(gen_ind,)`` in Tietze form under the representation_type of
        ``self`` from the data-file or via the ``GAP3`` interface

        INPUT:

        - ``gen_ind`` -- integer; index of a generator of the cubic Hecke
          algebra attached to ``self + 1``; negative values correspond to
          the according inverses

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: CHA3.<c1, c2> = algebras.CubicHecke(3)
            sage: MS = chmr.CubicHeckeMatrixSpace(CHA3)
            sage: MS._image_on_gen(1)
            [
                           [  b   0]  [  a   0]  [  a   0]
            [a], [c], [b], [b*c   c], [a*b   b], [a*c   c],
            <BLANKLINE>
            [        c         0         0]
            [b^2 + a*c         b         0]
            [        b         1         a]
            ]

            sage: CHA2 = CHA3.cubic_hecke_subalgebra()
            sage: MSreg = chmr.CubicHeckeMatrixSpace(CHA2, representation_type=CHA2.repr_type.RegularRight)
            sage: MSreg._image_on_gen(-1)
            [
            [     0      1 (-u)/w]
            [     0      0    1/w]
            [     1      0    v/w]
            ]
        """
        representation_type = self._representation_type
        original_base_ring = self._original_base_ring
        ch_algebra = self._cubic_hecke_algebra
        n = ch_algebra.strands()

        def invert_gen(matr):
            r"""
            Return the inverse matrix of generators.
            """
            cfs = ch_algebra.cubic_equation(as_coefficients=True, generic=True)
            fac = - 1 / cfs[0]
            cf0, cf1, cf2, cf3 = (original_base_ring(cf * fac) for cf in cfs)

            matri = cf1 * matr.parent().one()
            matri += cf2 * matr
            matri += cf3 * matr**2
            d1, d2 = matr.dimensions()
            return matrix(original_base_ring, d1, d2,
                          lambda i, j: original_base_ring(matri[i, j]))

        if n == 2:
            if representation_type.is_split():
                # Split representations for n == 2 are missing in CHEVIE and data files
                a, b, c = original_base_ring.gens()
                matrix_list = [matrix(1, 1, [a]), matrix(1, 1, [b]), matrix(1, 1, [c])]
                if gen_ind < 0:
                    matrix_list = [invert_gen(mat) for mat in matrix_list]
                return matrix_list

        num_rep = representation_type.number_of_representations(n)

        if representation_type == RepresentationType.SplitIrredChevie:
            rep_list = [ch_algebra._fetch_matrix_list_from_chevie(i + 1) for i in range(num_rep)]
            if gen_ind > 0:
                matrix_list = [rep[gen_ind - 1] for rep in rep_list]
            else:
                matrix_list = [invert_gen(rep[-gen_ind - 1]) for rep in rep_list]
        else:
            database = ch_algebra._database
            matrix_list = database.read_matrix_representation(representation_type, gen_ind, n, original_base_ring)
        return matrix_list

    @cached_method
    def _image_on_basis(self, basis_element):
        r"""
        Return the image of the given basis element of the cubic Hecke algebra
        in ``self``.

        INPUT:

        - ``basis_element`` -- a
          :class:`~sage.algebras.hecke_algebras.cubic_hecke_algebra.CubicHeckeElement`
          that is a monomial

        EXAMPLES::

            sage: import sage.algebras.hecke_algebras.cubic_hecke_matrix_rep as chmr
            sage: CHA3.<c1, c2> = algebras.CubicHecke(3)
            sage: MS = chmr.CubicHeckeMatrixSpace(CHA3, original=True)
            sage: MS._image_on_basis(c1)
            [        a         0         0         0         0         0         0         0         0         0         0         0]
            [        0         c         0         0         0         0         0         0         0         0         0         0]
            [        0         0         b         0         0         0         0         0         0         0         0         0]
            [        0         0         0         b         0         0         0         0         0         0         0         0]
            [        0         0         0       b*c         c         0         0         0         0         0         0         0]
            [        0         0         0         0         0         a         0         0         0         0         0         0]
            [        0         0         0         0         0       a*b         b         0         0         0         0         0]
            [        0         0         0         0         0         0         0         a         0         0         0         0]
            [        0         0         0         0         0         0         0       a*c         c         0         0         0]
            [        0         0         0         0         0         0         0         0         0         c         0         0]
            [        0         0         0         0         0         0         0         0         0 b^2 + a*c         b         0]
            [        0         0         0         0         0         0         0         0         0         b         1         a]
        """
        representation_type = self._representation_type
        ch_algebra = self._cubic_hecke_algebra
        filecache = ch_algebra._filecache

        original_base_ring = self._original_base_ring

        ele_Tietze = basis_element.Tietze()
        matrix_list = filecache.read_matrix_representation(representation_type, ele_Tietze, original_base_ring)
        if matrix_list is None:
            verbose('not in  memory %s (Tietze %s)' % (basis_element, ele_Tietze), level=2)
            if len(ele_Tietze) == 0:
                matrix_list = ch_algebra._create_matrix_list_for_one(representation_type)
            else:
                for gen_ind in ele_Tietze:
                    gen_matrix_list = self._image_on_gen(gen_ind)
                    if matrix_list is None:
                        matrix_list = list(gen_matrix_list)
                    else:
                        for i in range(len(matrix_list)):
                            matrix_list[i] *= gen_matrix_list[i]

            filecache.write_matrix_representation(representation_type, ele_Tietze, matrix_list)
            verbose('%s saved to memory' % basis_element, level=2)

        mat = block_diagonal_matrix(matrix_list, subdivide=self._subdivide, sparse=True)
        return self._specialize_matrix(mat)

    @cached_method
    def zero(self):
        r"""
        Return the zero element of ``self``.

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: m1   = c1.matrix()
            sage: m1rl = c1.matrix(representation_type = CHA2.repr_type.RegularLeft)
            sage: z   = m1.parent().zero()
            sage: zrl = m1rl.parent().zero()
            sage: matrix(z) == matrix(zrl), z.is_zero(), zrl.is_zero()
            (True, True, True)
            sage: z.block_diagonal_list()
            [[0], [0], [0]]
            sage: zrl.block_diagonal_list()
            [
            [0 0 0]
            [0 0 0]
            [0 0 0]
            ]
        """
        z = self.element_class(self, super().zero())
        z._cubic_hecke_element = self._cubic_hecke_algebra.zero()
        z.set_immutable()
        return z

    @cached_method
    def one(self):
        r"""
        Return the one element of ``self``.

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2)
            sage: m1   = c1.matrix()
            sage: m1rl = c1.matrix(representation_type = CHA2.repr_type.RegularLeft)
            sage: o   = m1.parent().one()
            sage: orl = m1rl.parent().one()
            sage: matrix(o) == matrix(orl), o.is_one(), orl.is_one()
            (True, True, True)
            sage: o.block_diagonal_list()
            [[1], [1], [1]]
            sage: orl.block_diagonal_list()
            [
            [1 0 0]
            [0 1 0]
            [0 0 1]
            ]
        """
        o = self.element_class(self, super().one())
        o._cubic_hecke_element = self._cubic_hecke_algebra.one()
        o.set_immutable()
        return o

    @cached_method
    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2, cubic_equation_roots=(2, 3, 5))
            sage: c1.matrix()
            [2 0 0]
            [0 3 0]
            [0 0 5]
            sage: _.parent()._an_element_()
            [ 94/3     0     0]
            [    0 187/3     0]
            [    0     0 373/3]
        """
        x = self._cubic_hecke_algebra.an_element()
        return self(x)

    @cached_method
    def some_elements(self):
        r"""
        Return a generator of elements of ``self``.

        EXAMPLES::

            sage: CHA2.<c1> = algebras.CubicHecke(2, cubic_equation_roots=(2, 3, 5))
            sage: M = c1.matrix(); M
            [2 0 0]
            [0 3 0]
            [0 0 5]
            sage: MS = M.parent()
            sage: MS.some_elements()
            (
            [ 94/3     0     0]
            [    0 187/3     0]
            [    0     0 373/3]
            )
            sage: MS.some_elements() == tuple(MS(x) for x in CHA2.some_elements())
            True
        """
        return tuple([self(x) for x in self._cubic_hecke_algebra.some_elements()])



# sage.doctest: needs sage.libs.flint
"""
Relation matrices for ambient modular symbols spaces

This file contains functions that are used by the various ambient modular
symbols classes to compute presentations of spaces in terms of generators and
relations, using the standard methods based on Manin symbols.
"""
# ****************************************************************************
#       Sage: Open Source Mathematical Software
#
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.matrix.matrix_space import MatrixSpace
from sage.misc.search import search
from sage.misc.verbose import verbose
from sage.modular.modsym.manin_symbol_list import ManinSymbolList
from sage.rings.rational_field import RationalField
from sage.categories.rings import Rings


SPARSE = True

# S = [0,-1; 1,0]
# T = [0,-1; 1,-1],
# T^2 = [-1, 1, -1, 0]
# I = [-1,0; 0,1]


######################################################################
# The following four functions are used to compute the quotient
# modulo the S, I, and T relations more efficiently that the generic
# code in the relation_matrix file:
#    modS_relations -- compute the S relations.
#    modI_quotient --  compute the I relations.
#    T_relation_matrix -- matrix whose echelon form gives
#                                 the quotient by 3-term T relations.
#    gens_to_basis_matrix -- compute echelon form of 3-term
#                                    relation matrix, and read off each
#                                    generator in terms of basis.
# These four functions are orchestrated in the function
#    compute_presentation
# which is defined below.  See the comment at the beginning
# of that function for an overall description of the algorithm.
######################################################################


def modS_relations(syms):
    r"""
    Compute quotient of Manin symbols by the S relations.

    Here S is the 2x2 matrix [0, -1; 1, 0].

    INPUT:

    - ``syms`` -- :class:`ManinSymbolList`

    OUTPUT:

    - ``rels`` -- set of pairs of pairs (j, s), where if
      mod[i] = (j,s), then x_i = s\*x_j (mod S relations)

    EXAMPLES::

        sage: from sage.modular.modsym.manin_symbol_list import ManinSymbolList_gamma0
        sage: from sage.modular.modsym.relation_matrix import modS_relations

    ::

        sage: syms = ManinSymbolList_gamma0(2, 4); syms
        Manin Symbol List of weight 4 for Gamma0(2)
        sage: modS_relations(syms)
        {((0, 1), (7, 1)),
         ((1, 1), (6, 1)),
         ((2, 1), (8, 1)),
         ((3, -1), (4, 1)),
         ((3, 1), (4, -1)),
         ((5, -1), (5, 1))}

    ::

        sage: syms = ManinSymbolList_gamma0(7, 2); syms
        Manin Symbol List of weight 2 for Gamma0(7)
        sage: modS_relations(syms)
        {((0, 1), (1, 1)), ((2, 1), (7, 1)), ((3, 1), (4, 1)), ((5, 1), (6, 1))}

    Next we do an example with Gamma1::

        sage: from sage.modular.modsym.manin_symbol_list import ManinSymbolList_gamma1
        sage: syms = ManinSymbolList_gamma1(3,2); syms
        Manin Symbol List of weight 2 for Gamma1(3)
        sage: modS_relations(syms)
        {((0, 1), (2, 1)),
         ((0, 1), (5, 1)),
         ((1, 1), (2, 1)),
         ((1, 1), (5, 1)),
         ((3, 1), (4, 1)),
         ((3, 1), (6, 1)),
         ((4, 1), (7, 1)),
         ((6, 1), (7, 1))}
    """
    if not isinstance(syms, ManinSymbolList):
        raise TypeError("syms must be a ManinSymbolList")
    tm = verbose()
    # We will fill in this set with the relations x_i + s*x_j = 0,
    # where the notation is as in _sparse_2term_quotient.
    rels = set()
    for i in range(len(syms)):
        j, s = syms.apply_S(i)
        assert j != -1
        if i < j:
            rels.add(((i, 1), (j, s)))
        else:
            rels.add(((j, s), (i, 1)))
    verbose("finished creating S relations", tm)
    return rels


def modI_relations(syms, sign):
    r"""
    Compute quotient of Manin symbols by the I relations.

    INPUT:

    - ``syms`` -- :class:`ManinSymbolList`

    - ``sign`` -- integer (either -1, 0, or 1)

    OUTPUT:

    - ``rels`` -- set of pairs of pairs (j, s), where if
      mod[i] = (j,s), then x_i = s\*x_j (mod S relations)

    EXAMPLES::

        sage: L = sage.modular.modsym.manin_symbol_list.ManinSymbolList_gamma1(4, 3)
        sage: sage.modular.modsym.relation_matrix.modI_relations(L, 1)
        {((0, 1), (0, -1)),
         ((1, 1), (1, -1)),
         ((2, 1), (8, -1)),
         ((3, 1), (9, -1)),
         ((4, 1), (10, -1)),
         ((5, 1), (11, -1)),
         ((6, 1), (6, -1)),
         ((7, 1), (7, -1)),
         ((8, 1), (2, -1)),
         ((9, 1), (3, -1)),
         ((10, 1), (4, -1)),
         ((11, 1), (5, -1)),
         ((12, 1), (12, 1)),
         ((13, 1), (13, 1)),
         ((14, 1), (20, 1)),
         ((15, 1), (21, 1)),
         ((16, 1), (22, 1)),
         ((17, 1), (23, 1)),
         ((18, 1), (18, 1)),
         ((19, 1), (19, 1)),
         ((20, 1), (14, 1)),
         ((21, 1), (15, 1)),
         ((22, 1), (16, 1)),
         ((23, 1), (17, 1))}

    .. warning::

       We quotient by the involution eta((u,v)) = (-u,v), which has
       the opposite sign as the involution in Merel's Springer LNM
       1585 paper! Thus our +1 eigenspace is his -1 eigenspace,
       etc. We do this for consistency with MAGMA.
    """
    tm = verbose()
    # We will fill in this set with the relations x_i - sign*s*x_j = 0,
    # where the notation is as in _sparse_2term_quotient.
    rels = set()
    for i in range(len(syms)):
        j, s = syms.apply_I(i)
        assert j != -1
        rels.add(((i, 1), (j, -sign * s)))
    verbose("finished creating I relations", tm)
    return rels


def T_relation_matrix_wtk_g0(syms, mod, field, sparse):
    r"""
    Compute a matrix whose echelon form gives the quotient by 3-term T
    relations. Despite the name, this is used for all modular symbols spaces
    (including those with character and those for `\Gamma_1` and `\Gamma_H`
    groups), not just `\Gamma_0`.

    INPUT:

    - ``syms`` -- :class:`ManinSymbolList`

    - ``mod`` -- list that gives quotient modulo some two-term relations, i.e.,
      the S relations, and if sign is nonzero, the I relations

    - ``field`` -- ``base_ring``

    - ``sparse`` -- boolean; whether to use sparse rather than dense
      linear algebra

    OUTPUT: a sparse matrix whose rows correspond to the reduction of
    the `T` relations modulo the `S` and `I` relations.

    EXAMPLES::

        sage: from sage.modular.modsym.relation_matrix import sparse_2term_quotient, T_relation_matrix_wtk_g0, modS_relations
        sage: L = sage.modular.modsym.manin_symbol_list.ManinSymbolList_gamma_h(GammaH(36, [17,19]), 2)
        sage: modS = sparse_2term_quotient(modS_relations(L), 216, QQ)
        sage: T_relation_matrix_wtk_g0(L, modS, QQ, False)
        72 x 216 dense matrix over Rational Field (use the '.str()' method to see the entries)
        sage: T_relation_matrix_wtk_g0(L, modS, GF(17), True)
        72 x 216 sparse matrix over Finite Field of size 17 (use the '.str()' method to see the entries)
    """
    tm = verbose()
    row = 0
    entries = {}
    already_seen = set()
    w = syms.weight()
    for i in range(len(syms)):
        if i in already_seen:
            continue
        iT_plus_iTT = syms.apply_T(i) + syms.apply_TT(i)
        j0, s0 = mod[i]
        v = {j0: s0}
        for j, s in iT_plus_iTT:
            if w == 2:
                already_seen.add(j)
            j0, s0 = mod[j]
            s0 = s * s0
            if j0 in v:
                v[j0] += s0
            else:
                v[j0] = s0
        for j0, vj0 in v.items():
            entries[(row, j0)] = vj0
        row += 1

    MAT = MatrixSpace(field, row, len(syms), sparse=True)
    R = MAT(entries)
    if not sparse:
        R = R.dense_matrix()
    verbose("finished (number of rows=%s)" % row, tm)
    return R


def gens_to_basis_matrix(syms, relation_matrix, mod, field, sparse):
    """
    Compute echelon form of 3-term relation matrix, and read off each
    generator in terms of basis.

    INPUT:

    - ``syms`` -- :class:`ManinSymbolList`

    - ``relation_matrix`` -- as output by
      ``__compute_T_relation_matrix(self, mod)``

    - ``mod`` -- quotient of modular symbols modulo the
      2-term S (and possibly I) relations

    - ``field`` -- base field

    - ``sparse`` -- boolean; whether or not matrix should be sparse

    OUTPUT:

    ``matrix`` -- a matrix whose `i`-th row expresses the Manin symbol
    generators in terms of a basis of Manin symbols (modulo the S, (possibly
    I,) and T rels). Note that the entries of the matrix need not be integers.

    - ``list`` -- integers `i`, such that the Manin symbols `x_i` are a basis

    EXAMPLES::

        sage: from sage.modular.modsym.relation_matrix import sparse_2term_quotient, T_relation_matrix_wtk_g0, gens_to_basis_matrix, modS_relations
        sage: L = sage.modular.modsym.manin_symbol_list.ManinSymbolList_gamma1(4, 3)
        sage: modS = sparse_2term_quotient(modS_relations(L), 24, GF(3))
        sage: gens_to_basis_matrix(L, T_relation_matrix_wtk_g0(L, modS, GF(3), 24), modS, GF(3), True)
        (24 x 2 sparse matrix over Finite Field of size 3, [13, 23])
    """
    from sage.structure.element import Matrix
    if not isinstance(relation_matrix, Matrix):
        raise TypeError("relation_matrix must be a matrix")
    if not isinstance(mod, list):
        raise TypeError("mod must be a list")

    verbose(str(relation_matrix.parent()))

    try:
        h = relation_matrix.height()
    except AttributeError:
        h = 9999999
    tm = verbose("putting relation matrix in echelon form (height = %s)" % h)
    if h < 10:
        A = relation_matrix.echelon_form(algorithm='multimodular',
                                         height_guess=1)
    else:
        A = relation_matrix.echelon_form()
    A.set_immutable()
    tm = verbose('finished echelon', tm)

    tm = verbose("Now creating gens --> basis mapping")

    basis_set = set(A.nonpivots())
    pivots = A.pivots()

    basis_mod2 = {j for j, c in mod if c != 0}

    basis_set = basis_set.intersection(basis_mod2)
    basis = sorted(basis_set)

    ONE = field(1)

    verbose("done doing setup", tm)

    tm = verbose("now forming quotient matrix")
    M = MatrixSpace(field, len(syms), len(basis), sparse=sparse)

    B = M(0)
    cols_index = {basis[i]: i for i in range(len(basis))}

    for i in basis_mod2:
        t, l = search(basis, i)
        if t:
            B[i, l] = ONE
        else:
            _, r = search(pivots, i)    # so pivots[r] = i
            # Set row i to -(row r of A), but where we only take
            # the non-pivot columns of A:
            B._set_row_to_negative_of_row_of_A_using_subset_of_columns(i, A, r, basis, cols_index)

    verbose("done making quotient matrix", tm)

    # The following is very fast (over Q at least).
    tm = verbose('now filling in the rest of the matrix')
    k = 0
    for i in range(len(mod)):
        j, s = mod[i]
        if j != i and s != 0:   # ignored in the above matrix
            k += 1
            B.set_row_to_multiple_of_row(i, j, s)
    verbose("set %s rows" % k)
    tm = verbose("time to fill in rest of matrix", tm)

    return B, basis


def compute_presentation(syms, sign, field, sparse=None):
    r"""
    Compute the presentation for self, as a quotient of Manin symbols
    modulo relations.

    INPUT:

    - ``syms`` -- :class:`ManinSymbolList`

    - ``sign`` -- integer (-1, 0, 1)

    - ``field`` -- a field

    OUTPUT:

    - sparse matrix whose rows give each generator
      in terms of a basis for the quotient

    - list of integers that give the basis for the
      quotient

    - mod: list where mod[i]=(j,s) means that x_i
      = s\*x_j modulo the 2-term S (and possibly I) relations.


    ALGORITHM:

    #. Let `S = [0,-1; 1,0], T = [0,-1; 1,-1]`, and
       `I = [-1,0; 0,1]`.

    #. Let `x_0,\ldots, x_{n-1}` by a list of all
       non-equivalent Manin symbols.

    #. Form quotient by 2-term S and (possibly) I relations.

    #. Create a sparse matrix `A` with `m` columns,
       whose rows encode the relations

       .. MATH::

                          [x_i] + [x_i T] + [x_i T^2] = 0.


       There are about n such rows. The number of nonzero entries per row
       is at most 3\*(k-1). Note that we must include rows for *all* i,
       since even if `[x_i] = [x_j]`, it need not be the case
       that `[x_i T] = [x_j T]`, since `S` and
       `T` do not commute. However, in many cases we have an a
       priori formula for the dimension of the quotient by all these
       relations, so we can omit many relations and just check that there
       are enough at the end--if there aren't, we add in more.

    #. Compute the reduced row echelon form of `A` using sparse
       Gaussian elimination.

    #. Use what we've done above to read off a sparse matrix R that
       uniquely expresses each of the n Manin symbols in terms of a subset
       of Manin symbols, modulo the relations. This subset of Manin
       symbols is a basis for the quotient by the relations.

    EXAMPLES::

        sage: L = sage.modular.modsym.manin_symbol_list.ManinSymbolList_gamma0(8,2)
        sage: sage.modular.modsym.relation_matrix.compute_presentation(L, 1, GF(9,'a'), True)
        (
        [2 0 0]
        [1 0 0]
        [0 0 0]
        [0 2 0]
        [0 0 0]
        [0 0 2]
        [0 0 0]
        [0 2 0]
        [0 0 0]
        [0 1 0]
        [0 1 0]
        [0 0 1], [1, 9, 11], [(1, 2), (1, 1), (0, 0), (9, 2), (0, 0), (11, 2), (0, 0), (9, 2), (0, 0), (9, 1), (9, 1), (11, 1)]
        )
    """
    if sparse is None:
        if syms.weight() >= 6:
            sparse = False
        else:
            sparse = True
    R, mod = relation_matrix_wtk_g0(syms, sign, field, sparse)
    B, basis = gens_to_basis_matrix(syms, R, mod, field, sparse)
    return B, basis, mod


def relation_matrix_wtk_g0(syms, sign, field, sparse):
    r"""
    Compute the matrix of relations. Despite the name, this is used for all
    spaces (not just for Gamma0). For a description of the algorithm, see the
    docstring for ``compute_presentation``.

    INPUT:

    - ``syms`` -- :class:`ManinSymbolList`

    - ``sign`` -- integer (0, 1 or -1)

    - ``field`` -- the base field (non-field base rings not supported at present)

    - ``sparse`` -- boolean; whether to use sparse arithmetic

    Note that ManinSymbolList objects already have a specific weight, so there
    is no need for an extra ``weight`` parameter.

    OUTPUT: a pair (R, mod) where

    - R is a matrix as output by ``T_relation_matrix_wtk_g0``

    - mod is a set of 2-term relations as output by ``sparse_2term_quotient``

    EXAMPLES::

        sage: L = sage.modular.modsym.manin_symbol_list.ManinSymbolList_gamma0(8,2)
        sage: A = sage.modular.modsym.relation_matrix.relation_matrix_wtk_g0(L, 0, GF(2), True); A
        (
        [0 0 0 0 0 0 0 0 1 0 0 0]
        [0 0 0 0 0 0 0 0 1 1 1 0]
        [0 0 0 0 0 0 1 0 0 1 1 0]
        [0 0 0 0 0 0 1 0 0 0 0 0], [(1, 1), (1, 1), (8, 1), (10, 1), (6, 1), (11, 1), (6, 1), (9, 1), (8, 1), (9, 1), (10, 1), (11, 1)]
        )
        sage: A[0].is_sparse()
        True
    """
    rels = modS_relations(syms)
    if sign != 0:
        # Let rels = rels union I relations.
        rels.update(modI_relations(syms, sign))

    rels = sorted(rels)
    # required for stability of doctests with python3

    if syms._apply_S_only_0pm1() and isinstance(field, RationalField):
        from . import relation_matrix_pyx
        mod = relation_matrix_pyx.sparse_2term_quotient_only_pm1(rels, len(syms))
    else:
        mod = sparse_2term_quotient(rels, len(syms), field)

    R = T_relation_matrix_wtk_g0(syms, mod, field, sparse)
    return R, mod


def sparse_2term_quotient(rels, n, F):
    r"""
    Perform Sparse Gauss elimination on a matrix all of whose columns
    have at most 2 nonzero entries. We use an obvious algorithm, which
    runs fast enough. (Typically making the list of relations takes
    more time than computing this quotient.) This algorithm is more
    subtle than just "identify symbols in pairs", since complicated
    relations can cause generators to surprisingly equal 0.

    INPUT:

    - ``rels`` -- iterable made of pairs ((i,s), (j,t)). The pair
      represents the relation `s x_i + t x_j = 0`, where the `i, j` must
      be Python int's.

    - ``n`` -- integer, the `x_i` are `x_0, \ldots, x_{n-1}`

    - ``F`` -- base field

    OUTPUT:

    ``mod`` -- list such that ``mod[i] = (j,s)``, which means that `x_i` is
    equivalent to `s x_j`, where the `x_j` are a basis for the quotient.

    EXAMPLES: We quotient out by the relations

    .. MATH::

        3*x0 - x1 = 0,\qquad  x1 + x3 = 0,\qquad   x2 + x3 = 0,\qquad  x4 - x5 = 0

    to get::

        sage: rels = [((int(0),3), (int(1),-1)), ((int(1),1), (int(3),1)), ((int(2),1),(int(3),1)), ((int(4),1),(int(5),-1))]
        sage: n = 6
        sage: from sage.modular.modsym.relation_matrix import sparse_2term_quotient
        sage: sparse_2term_quotient(rels, n, QQ)
        [(3, -1/3), (3, -1), (3, -1), (3, 1), (5, 1), (5, 1)]
    """
    n = int(n)
    if F not in Rings():
        raise TypeError("F must be a ring")

    tm = verbose("Starting sparse 2-term quotient...")
    free = list(range(n))
    ONE = F.one()
    ZERO = F.zero()
    coef = [ONE for i in range(n)]
    related_to_me = [[] for i in range(n)]
    for v0, v1 in sorted(rels):
        c0 = coef[v0[0]] * F(v0[1])
        c1 = coef[v1[0]] * F(v1[1])

        # Mod out by the following relation:
        #
        #    c0*free[v0[0]] + c1*free[v1[0]] = 0.
        #
        die = None
        if c0 == ZERO and c1 == ZERO:
            pass
        elif c0 == ZERO and c1 != ZERO:  # free[v1[0]] --> 0
            die = free[v1[0]]
        elif c1 == ZERO and c0 != ZERO:
            die = free[v0[0]]
        elif free[v0[0]] == free[v1[0]]:
            if c0 + c1 != 0:
                # all xi equal to free[v0[0]] must now equal to zero.
                die = free[v0[0]]
        else:  # x1 = -c1/c0 * x2.
            x = free[v0[0]]
            free[x] = free[v1[0]]
            coef[x] = -c1 / c0
            for i in related_to_me[x]:
                free[i] = free[x]
                coef[i] *= coef[x]
                related_to_me[free[v1[0]]].append(i)
            related_to_me[free[v1[0]]].append(x)
        if die is not None:
            for i in related_to_me[die]:
                free[i] = 0
                coef[i] = ZERO
            free[die] = 0
            coef[die] = ZERO

    mod = [(free[i], coef[i]) for i in range(len(free))]
    verbose("finished", tm)
    return mod



r"""
Interface to the Gnuplot interpreter
"""

# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import os
import time
from sage.structure.sage_object import SageObject


class Gnuplot(SageObject):
    """
    Interface to the Gnuplot interpreter.
    """
    def _quit_string(self):
        return 'quit'

    def gnuplot(self):
        try:
            return self._gnuplot
        except AttributeError:
            try:
                import Gnuplot as GP
                self._gnuplot = GP.Gnuplot()
                return self._gnuplot
            except ImportError:
                raise RuntimeError("Install the gnuplotpy Python module.")

    def __call__(self, line):
        return self.gnuplot()(line)

    def _eval_line(self, line, *args, **kwds):
        self(line)
        return ''

    def _repr_(self):
        """
        TESTS::

            sage: gnuplot               # indirect doctests
            Interface to Gnuplot
        """
        return "Interface to Gnuplot"

    def plot(self, cmd, file=None, verbose=True, reset=True):
        r"""
        Draw the plot described by cmd, and possibly also save to an eps or
        png file.

        INPUT:

        - ``cmd`` -- string

        - ``file`` -- string (default: ``None``); if specified save
          plot to given file, which may be either an eps (default) or png file

        - ``verbose`` -- print some info

        - ``reset`` -- true; reset gnuplot before making graph

        OUTPUT: displays graph

        .. NOTE::

           Note that ``^`` s  are replaced by ``**`` s before being passed to gnuplot.
        """
        if reset:
            self('reset')
        self('set terminal x11')
        cmd = cmd.replace('^', '**')
        self(cmd)
        if file is not None:
            if file[-4:] == '.png':
                self('set terminal png medium')
            else:
                if file[-4:] != '.eps':
                    file += '.eps'
                self('set terminal postscript eps enhanced')
            # self("set output '%s'"%file)
            tmp = 'gnuplot_tmp%s' % file[-4:]
            self("set output '%s'" % tmp)
            print("Saving plot to %s" % file)
            self(cmd)
            time.sleep(0.1)
            os.system('mv %s %s 2>/dev/null' % (tmp, file))
            time.sleep(0.1)
        self('set terminal x11')

    def plot3d(self, f, xmin=-1, xmax=1, ymin=-1, ymax=1, zmin=-1, zmax=1,
               title=None,
               samples=25, isosamples=20, xlabel='x', ylabel='y',
               interact=True):
        if title is None:
            title = str(f)
        f = f.replace('^', '**')
        cmd = """
        set xlabel "%s"
        set ylabel "%s"
        set key top
        set border 4095
        set xrange [%s:%s]
        set yrange [%s:%s]
        set samples %s
        set isosamples %s

        set title "%s"
        set pm3d; set palette
        #show pm3d
        #show palette
        splot %s
        """ % (xlabel, ylabel,
             xmin, xmax, ymin, ymax,  # zmin, zmax,
             samples, isosamples,
             title, f)
        if interact:
            self.interact(cmd)
        else:
            self(cmd)

    def plot3d_parametric(self, f='cos(u)*(3 + v*cos(u/2)), sin(u)*(3 + v*cos(u/2)), v*sin(u/2)',
                          range1='[u=-pi:pi]',
                          range2='[v=-0.2:0.2]', samples=50, title=None,
                          interact=True):
        r"""
        Draw a parametric 3d surface and rotate it interactively.

        INPUT:

        - ``f`` -- string; a function of two variables, e.g.,
          ``'cos(u)\*(3 + v\*cos(u/2)), sin(u)\*(3 + v\*cos(u/2)), v\*sin(u/2)'``

        - ``range1`` -- string; range of values for one
          variable, e.g., ``'[u=-pi:pi]'``

        - ``range2`` -- string; range of values for another
          variable, e.g., ``'[v=-0.2:0.2]'``

        - ``samples`` -- integer; number of sample points to use

        - ``title`` -- string; title of the graph

        EXAMPLES::

            sage: gnuplot.plot3d_parametric('v^2*sin(u), v*cos(u), v*(1-v)')   # optional - gnuplot, not tested (since something pops up)
        """
        if title is None:
            title = str(f)
        cmd = """
        set key top
        set border 4095
        set samples %s

        set title "%s"
        set pm3d; set palette; set parametric
        splot %s %s %s
        """ % (samples, title, range1, range2, f)
        cmd = cmd.replace('^', '**')
        if interact:
            self.interact(cmd)
        else:
            self(cmd)

    def interact(self, cmd):
        import tempfile
        with tempfile.NamedTemporaryFile(mode='w+t') as f:
            f.write(cmd + '\n pause -1 "Press return to continue (no further rotation possible)"')
            os.system(f'gnuplot -persist {f.name}')

    def console(self):
        gnuplot_console()


# An instance
gnuplot = Gnuplot()


def gnuplot_console():
    from sage.repl.rich_output.display_manager import get_display_manager
    if not get_display_manager().is_in_terminal():
        raise RuntimeError('Can use the console only in the terminal. Try %%gnuplot magics instead.')
    os.system('gnuplot')



# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Grossman-Larson Hopf algebras

AUTHORS:

- Frédéric Chapoton (2017)
"""

# ****************************************************************************
#       Copyright (C) 2017 Frédéric Chapoton
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.hopf_algebras import HopfAlgebras
from sage.combinat.free_module import CombinatorialFreeModule
from sage.combinat.words.alphabet import Alphabet
from sage.combinat.rooted_tree import (RootedTrees, RootedTree,
                                       LabelledRootedTrees,
                                       LabelledRootedTree)
from sage.categories.rings import Rings
from sage.sets.family import Family
from sage.rings.integer_ring import ZZ
from itertools import combinations, product

# we use a fixed special symbol for the fake root
ROOT = '#'


class GrossmanLarsonAlgebra(CombinatorialFreeModule):
    r"""
    The Grossman-Larson Hopf Algebra.

    The Grossman-Larson Hopf Algebras are Hopf algebras with a basis
    indexed by forests of decorated rooted trees. They are the
    universal enveloping algebras of free pre-Lie algebras, seen
    as Lie algebras.

    The Grossman-Larson Hopf algebra on a given set `E` has an
    explicit description using rooted forests. The underlying vector
    space has a basis indexed by finite rooted forests endowed with a
    map from their vertices to `E` (called the "labeling").
    In this basis, the product of two
    (decorated) rooted forests `S * T` is a sum over all maps from
    the set of roots of `T` to the union of a singleton `\{\#\}` and
    the set of vertices of `S`. Given such a map, one defines a new
    forest as follows. Starting from the disjoint union of all rooted trees
    of `S` and `T`, one adds an edge from every root of `T` to its
    image when this image is not the fake vertex labelled ``#``.
    The coproduct sends a rooted forest `T` to the sum of all tensors
    `T_1 \otimes T_2` obtained by splitting the connected components
    of `T` into two subsets and letting `T_1` be the forest formed
    by the first subset and `T_2` the forest formed by the second.
    This yields a connected graded Hopf algebra (the degree of a
    forest is its number of vertices).

    See [Pana2002]_ (Section 2) and [GroLar1]_.
    (Note that both references use rooted trees rather than rooted
    forests, so think of each rooted forest grafted onto a new root.
    Also, the product is reversed, so they are defining the opposite
    algebra structure.)

    .. WARNING::

        For technical reasons, instead of using forests as labels for
        the basis, we use rooted trees. Their root vertex should be
        considered as a fake vertex. This fake root vertex is labelled
        ``'#'`` when labels are present.

    EXAMPLES::

        sage: G = algebras.GrossmanLarson(QQ, 'xy')
        sage: x, y = G.single_vertex_all()
        sage: ascii_art(x*y)
        B  + B
         #      #_
         |     / /
         x    x y
         |
         y

        sage: ascii_art(x*x*x)
        B  + B     + 3*B     + B
         #      #         #_      _#__
         |      |        / /     / / /
         x      x_      x x     x x x
         |     / /        |
         x    x x         x
         |
         x

    The Grossman-Larson algebra is associative::

        sage: z = x * y
        sage: x * (y * z) == (x * y) * z
        True

    It is not commutative::

        sage: x * y == y * x
        False

    When ``None`` is given as input, unlabelled forests are used instead;
    this corresponds to a `1`-element set `E`::

        sage: G = algebras.GrossmanLarson(QQ, None)
        sage: x = G.single_vertex_all()[0]
        sage: ascii_art(x*x)
        B  + B
         o      o_
         |     / /
         o    o o
         |
         o

    .. NOTE::

        Variables names can be ``None``, a list of strings, a string
        or an integer. When ``None`` is given, unlabelled rooted
        forests are used. When a single string is given, each letter is taken
        as a variable. See
        :func:`sage.combinat.words.alphabet.build_alphabet`.

    .. WARNING::

        Beware that the underlying combinatorial free module is based
        either on ``RootedTrees`` or on ``LabelledRootedTrees``, with no
        restriction on the labellings. This means that all code calling
        the :meth:`basis` method would not give meaningful results, since
        :meth:`basis` returns many "chaff" elements that do not belong to
        the algebra.

    REFERENCES:

    - [Pana2002]_

    - [GroLar1]_
    """
    @staticmethod
    def __classcall_private__(cls, R, names=None):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: F1 = algebras.GrossmanLarson(QQ, 'xyz')
            sage: F2 = algebras.GrossmanLarson(QQ, ['x','y','z'])
            sage: F3 = algebras.GrossmanLarson(QQ, Alphabet('xyz'))
            sage: F1 is F2 and F1 is F3
            True
        """
        if names is not None:
            if names not in ZZ and ',' in names:
                names = [u for u in names if u != ',']
            names = Alphabet(names)

        if R not in Rings():
            raise TypeError("argument R must be a ring")
        return super().__classcall__(cls, R, names)

    def __init__(self, R, names=None):
        """
        Initialize ``self``.

        TESTS::

            sage: A = algebras.GrossmanLarson(QQ, '@'); A
            Grossman-Larson Hopf algebra on one generator ['@']
            over Rational Field
            sage: TestSuite(A).run() # long time

            sage: F = algebras.GrossmanLarson(QQ, 'xy')
            sage: TestSuite(F).run() # long time

            sage: A = algebras.GrossmanLarson(QQ, None); A
            Grossman-Larson Hopf algebra on one generator ['o'] over
            Rational Field

            sage: F = algebras.GrossmanLarson(QQ, ['x','y']); F
            Grossman-Larson Hopf algebra on 2 generators ['x', 'y']
            over Rational Field

            sage: A = algebras.GrossmanLarson(QQ, []); A
            Grossman-Larson Hopf algebra on 0 generators [] over
            Rational Field
        """
        if names is None:
            Trees = RootedTrees()
            key = RootedTree.sort_key
            self._alphabet = Alphabet(['o'])
        else:
            Trees = LabelledRootedTrees()
            key = LabelledRootedTree.sort_key
            self._alphabet = names
        # Here one would need LabelledRootedTrees(names)
        # so that one can restrict the labels to some fixed set

        cat = HopfAlgebras(R).WithBasis().Graded()
        CombinatorialFreeModule.__init__(self, R, Trees,
                                         latex_prefix='',
                                         sorting_key=key,
                                         category=cat)

    def variable_names(self):
        r"""
        Return the names of the variables.

        This returns the set `E` (as a family).

        EXAMPLES::

            sage: R = algebras.GrossmanLarson(QQ, 'xy')
            sage: R.variable_names()
            {'x', 'y'}

            sage: R = algebras.GrossmanLarson(QQ, ['a','b'])
            sage: R.variable_names()
            {'a', 'b'}

            sage: R = algebras.GrossmanLarson(QQ, 2)
            sage: R.variable_names()
            {0, 1}

            sage: R = algebras.GrossmanLarson(QQ, None)
            sage: R.variable_names()
            {'o'}
        """
        return self._alphabet

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: algebras.GrossmanLarson(QQ, '@')  # indirect doctest
            Grossman-Larson Hopf algebra on one generator ['@'] over Rational Field
            sage: algebras.GrossmanLarson(QQ, None)  # indirect doctest
            Grossman-Larson Hopf algebra on one generator ['o'] over Rational Field
            sage: algebras.GrossmanLarson(QQ, ['a','b'])
            Grossman-Larson Hopf algebra on 2 generators ['a', 'b'] over Rational Field
        """
        n = len(self.single_vertex_all())
        if n == 1:
            gen = "one generator"
        else:
            gen = "{} generators".format(n)
        s = "Grossman-Larson Hopf algebra on {} {} over {}"
        try:
            return s.format(gen, self._alphabet.list(), self.base_ring())
        except NotImplementedError:
            return s.format(gen, self._alphabet, self.base_ring())

    def single_vertex(self, i):
        r"""
        Return the ``i``-th rooted forest with one vertex.

        This is the rooted forest with just one vertex, labelled by the
        ``i``-th element of the label list.

        .. SEEALSO:: :meth:`single_vertex_all`.

        INPUT:

        - ``i`` -- nonnegative integer

        EXAMPLES::

            sage: F = algebras.GrossmanLarson(ZZ, 'xyz')
            sage: F.single_vertex(0)
            B[#[x[]]]

            sage: F.single_vertex(4)
            Traceback (most recent call last):
            ...
            IndexError: argument i (= 4) must be between 0 and 2
        """
        G = self.single_vertex_all()
        n = len(G)
        if i < 0 or not i < n:
            m = "argument i (= {}) must be between 0 and {}".format(i, n - 1)
            raise IndexError(m)
        return G[i]

    def single_vertex_all(self):
        """
        Return the rooted forests with one vertex in ``self``.

        They freely generate the Lie algebra of primitive elements
        as a pre-Lie algebra.

        .. SEEALSO:: :meth:`single_vertex`.

        EXAMPLES::

            sage: A = algebras.GrossmanLarson(ZZ, 'fgh')
            sage: A.single_vertex_all()
            (B[#[f[]]], B[#[g[]]], B[#[h[]]])

            sage: A = algebras.GrossmanLarson(QQ, ['x1','x2'])
            sage: A.single_vertex_all()
            (B[#[x1[]]], B[#[x2[]]])

            sage: A = algebras.GrossmanLarson(ZZ, None)
            sage: A.single_vertex_all()
            (B[[[]]],)
        """
        Trees = self.basis().keys()
        return tuple(Family(self._alphabet,
                            lambda a: self.monomial(Trees([Trees([], a)], ROOT))))

    def _first_ngens(self, n):
        """
        Return the first generators.

        EXAMPLES::

            sage: A = algebras.GrossmanLarson(QQ, ['x1','x2'])
            sage: A._first_ngens(2)
            (B[#[x1[]]], B[#[x2[]]])

            sage: A = algebras.GrossmanLarson(ZZ, None)
            sage: A._first_ngens(1)
            (B[[[]]],)
        """
        return self.single_vertex_all()[:n]

    def change_ring(self, R):
        """
        Return the Grossman-Larson algebra in the same variables over `R`.

        INPUT:

        - ``R`` -- a ring

        EXAMPLES::

            sage: A = algebras.GrossmanLarson(ZZ, 'fgh')
            sage: A.change_ring(QQ)
            Grossman-Larson Hopf algebra on 3 generators ['f', 'g', 'h']
            over Rational Field
        """
        return GrossmanLarsonAlgebra(R, names=self.variable_names())

    def degree_on_basis(self, t):
        """
        Return the degree of a rooted forest in the Grossman-Larson algebra.

        This is the total number of vertices of the forest.

        EXAMPLES::

            sage: A = algebras.GrossmanLarson(QQ, '@')
            sage: RT = A.basis().keys()
            sage: A.degree_on_basis(RT([RT([])]))
            1
        """
        return t.number_of_nodes() - 1

    def _an_element_(self):
        """
        Return an element of ``self``.

        EXAMPLES::

            sage: A = algebras.GrossmanLarson(QQ, 'xy')
            sage: A.an_element()
            B[#[x[]]] + 2*B[#[x[x[]]]] + 2*B[#[x[], x[]]]
        """
        o = self.single_vertex(0)
        return o + 2 * o * o

    def some_elements(self):
        """
        Return some elements of the Grossman-Larson Hopf algebra.

        EXAMPLES::

            sage: A = algebras.GrossmanLarson(QQ, None)
            sage: A.some_elements()
            [B[[[]]], B[[]] + B[[[[]]]] + B[[[], []]],
            4*B[[[[]]]] + 4*B[[[], []]]]

        With several generators::

            sage: A = algebras.GrossmanLarson(QQ, 'xy')
            sage: A.some_elements()
            [B[#[x[]]],
             B[#[]] + B[#[x[x[]]]] + B[#[x[], x[]]],
             B[#[x[x[]]]] + 3*B[#[x[y[]]]] + B[#[x[], x[]]] + 3*B[#[x[], y[]]]]
        """
        o = self.single_vertex(0)
        o1 = self.single_vertex_all()[-1]
        x = o * o
        y = o * o1
        return [o, 1 + x, x + 3 * y]

    def product_on_basis(self, x, y):
        """
        Return the product of two forests `x` and `y`.

        This is the sum over all possible ways for the components
        of the forest `y` to either fall side-by-side with components
        of `x` or be grafted on a vertex of `x`.

        EXAMPLES::

            sage: A = algebras.GrossmanLarson(QQ, None)
            sage: RT = A.basis().keys()
            sage: x = RT([RT([])])
            sage: A.product_on_basis(x, x)
            B[[[[]]]] + B[[[], []]]

        Check that the product is the correct one::

            sage: A = algebras.GrossmanLarson(QQ, 'uv')
            sage: RT = A.basis().keys()
            sage: Tu = RT([RT([],'u')],'#')
            sage: Tv = RT([RT([],'v')],'#')
            sage: A.product_on_basis(Tu, Tv)
            B[#[u[v[]]]] + B[#[u[], v[]]]
        """
        return self.sum(self.basis()[x.single_graft(y, graftingFunction)]
                        for graftingFunction in
                        product(list(x.paths()), repeat=len(y)))

    def one_basis(self):
        """
        Return the empty rooted forest.

        EXAMPLES::

            sage: A = algebras.GrossmanLarson(QQ, 'ab')
            sage: A.one_basis()
            #[]

            sage: A = algebras.GrossmanLarson(QQ, None)
            sage: A.one_basis()
            []
        """
        Trees = self.basis().keys()
        return Trees([], ROOT)

    def coproduct_on_basis(self, x):
        """
        Return the coproduct of a forest.

        EXAMPLES::

            sage: G = algebras.GrossmanLarson(QQ,2)
            sage: x, y = G.single_vertex_all()
            sage: ascii_art(G.coproduct(x))  # indirect doctest
            1 # B  + B  # 1
                 #    #
                 |    |
                 0    0

            sage: Delta_xy = G.coproduct(y*x)
            sage: ascii_art(Delta_xy)  # random indirect doctest
            1 # B     + 1 # B  + B  # B  + B     # 1 + B  # B  + B  # 1
                   #_        #    #    #      #_        #    #    #
                  / /        |    |    |     / /        |    |    |
                 0 1         1    0    1    0 1         1    0    1
                             |                                    |
                             0                                    0

        TESTS::

            sage: Delta_xy.coefficients()
            [1, 1, 1, 1, 1, 1]
            sage: sortkey = G.print_options()['sorting_key']
            sage: doublekey = lambda tt: (sortkey(tt[0]), sortkey(tt[1]))
            sage: sorted(Delta_xy.monomial_coefficients(), key=doublekey)
            [(#[], #[1[0[]]]),
             (#[], #[0[], 1[]]),
             (#[0[]], #[1[]]),
             (#[1[]], #[0[]]),
             (#[1[0[]]], #[]),
             (#[0[], 1[]], #[])]
        """
        B = self.basis()
        Trees = B.keys()
        subtrees = list(x)
        num_subtrees = len(subtrees)
        indx = list(range(num_subtrees))
        return sum(B[Trees([subtrees[i] for i in S], ROOT)].tensor(
                   B[Trees([subtrees[i] for i in indx if i not in S], ROOT)])
                   for k in range(num_subtrees + 1)
                   for S in combinations(indx, k))

    def counit_on_basis(self, x):
        """
        Return the counit on a basis element.

        This is zero unless the forest `x` is empty.

        EXAMPLES::

            sage: A = algebras.GrossmanLarson(QQ, 'xy')
            sage: RT = A.basis().keys()
            sage: x = RT([RT([],'x')],'#')
            sage: A.counit_on_basis(x)
            0
            sage: A.counit_on_basis(RT([],'#'))
            1
        """
        if x.number_of_nodes() == 1:
            return self.base_ring().one()
        return self.base_ring().zero()

    def antipode_on_basis(self, x):
        """
        Return the antipode of a forest.

        EXAMPLES::

            sage: G = algebras.GrossmanLarson(QQ,2)
            sage: x, y = G.single_vertex_all()
            sage: G.antipode(x)  # indirect doctest
            -B[#[0[]]]

            sage: G.antipode(y*x)  # indirect doctest
            B[#[0[1[]]]] + B[#[0[], 1[]]]
        """
        B = self.basis()
        Trees = B.keys()
        subtrees = list(x)
        if not subtrees:
            return self.one()
        num_subtrees = len(subtrees)
        indx = list(range(num_subtrees))
        return sum(- self.antipode_on_basis(Trees([subtrees[i] for i in S], ROOT))
                   * B[Trees([subtrees[i] for i in indx if i not in S], ROOT)]
                   for k in range(num_subtrees)
                   for S in combinations(indx, k))

    def _element_constructor_(self, x):
        r"""
        Convert ``x`` into ``self``.

        EXAMPLES::

            sage: R = algebras.GrossmanLarson(QQ, 'xy')
            sage: x, y = R.single_vertex_all()
            sage: R(x)
            B[#[x[]]]
            sage: R(x+4*y)
            B[#[x[]]] + 4*B[#[y[]]]

            sage: Trees = R.basis().keys()
            sage: R(Trees([],'#'))
            B[#[]]

            sage: D = algebras.GrossmanLarson(ZZ, 'xy')
            sage: X, Y = D.single_vertex_all()
            sage: R(X-Y).parent()
            Grossman-Larson Hopf algebra on 2 generators ['x', 'y'] over Rational Field

        TESTS::

            sage: Trees = R.basis().keys()
            sage: R(Trees([],'x'))
            Traceback (most recent call last):
            ...
            ValueError: incorrect root label

            sage: R.<x,y> = algebras.GrossmanLarson(QQ)
            sage: R(x) is x
            True
            sage: S.<z> = algebras.GrossmanLarson(GF(3))
            sage: R(z)
            Traceback (most recent call last):
            ...
            TypeError: not able to convert this to this algebra
        """
        if (isinstance(x, (RootedTree, LabelledRootedTree))
                and x in self.basis().keys()):
            if hasattr(x, 'label') and x.label() != ROOT:
                raise ValueError('incorrect root label')
            return self.monomial(x)
        try:
            P = x.parent()
            if isinstance(P, GrossmanLarsonAlgebra):
                if P is self:
                    return x
                if self._coerce_map_from_(P):
                    return self.element_class(self, x.monomial_coefficients())
        except AttributeError:
            raise TypeError('not able to convert this to this algebra')
        else:
            raise TypeError('not able to convert this to this algebra')
        # Ok, not an element (or should not be viewed as one).

    def _coerce_map_from_(self, R):
        r"""
        Return ``True`` if there is a coercion from ``R`` into ``self``
        and ``False`` otherwise.

        The things that coerce into ``self`` are

        - Grossman-Larson Hopf algebras whose set `E` of labels is
          a subset of the corresponding set of ``self``, and whose base
          ring has a coercion map into ``self.base_ring()``

        EXAMPLES::

            sage: F = algebras.GrossmanLarson(GF(7), 'xyz'); F
            Grossman-Larson Hopf algebra on 3 generators ['x', 'y', 'z']
             over Finite Field of size 7

        Elements of the Grossman-Larson Hopf algebra canonically coerce in::

            sage: x, y, z = F.single_vertex_all()
            sage: F.coerce(x+y) == x+y
            True

        The Grossman-Larson Hopf algebra over `\ZZ` on `x, y, z`
        coerces in, since `\ZZ` coerces to `\GF{7}`::

            sage: G = algebras.GrossmanLarson(ZZ, 'xyz')
            sage: Gx,Gy,Gz = G.single_vertex_all()
            sage: z = F.coerce(Gx+Gy); z
            B[#[x[]]] + B[#[y[]]]
            sage: z.parent() is F
            True

        However, `\GF{7}` does not coerce to `\ZZ`, so the Grossman-Larson
        algebra over `\GF{7}` does not coerce to the one over `\ZZ`::

            sage: G.coerce(y)
            Traceback (most recent call last):
            ...
            TypeError: no canonical coercion from Grossman-Larson Hopf algebra
             on 3 generators ['x', 'y', 'z'] over Finite Field of size
             7 to Grossman-Larson Hopf algebra on 3 generators ['x', 'y', 'z']
             over Integer Ring

        TESTS::

            sage: F = algebras.GrossmanLarson(ZZ, 'xyz')
            sage: G = algebras.GrossmanLarson(QQ, 'xyz')
            sage: H = algebras.GrossmanLarson(ZZ, 'y')
            sage: F._coerce_map_from_(G)
            False
            sage: G._coerce_map_from_(F)
            True
            sage: F._coerce_map_from_(H)
            True
            sage: F._coerce_map_from_(QQ)
            False
            sage: G._coerce_map_from_(QQ)
            False
            sage: F.has_coerce_map_from(PolynomialRing(ZZ, 3, 'x,y,z'))
            False
        """
        # Grossman-Larson algebras containing the same variables
        # over any base that coerces in:
        if isinstance(R, GrossmanLarsonAlgebra):
            if all(x in self.variable_names() for x in R.variable_names()):
                if self.base_ring().has_coerce_map_from(R.base_ring()):
                    return True
        return False



"""
Generic LibGAP-based Group

This is useful if you need to use a GAP group implementation in Sage
that does not have a dedicated Sage interface.

If you want to implement your own group class, you should not derive
from this but directly from
:class:`~sage.groups.libgap_wrapper.ParentLibGAP`.

EXAMPLES::

    sage: F.<a,b> = FreeGroup()
    sage: G_gap = libgap.Group([ (a*b^2).gap() ])
    sage: from sage.groups.libgap_group import GroupLibGAP
    sage: G = GroupLibGAP(G_gap);  G
    Group([ a*b^2 ])
    sage: type(G)
    <class 'sage.groups.libgap_group.GroupLibGAP_with_category'>
    sage: G.gens()
    (a*b^2,)
"""

##############################################################################
#       Copyright (C) 2013 Volker Braun <vbraun.name@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
##############################################################################


from sage.groups.group import Group
from sage.groups.libgap_wrapper import ParentLibGAP, ElementLibGAP
from sage.groups.libgap_mixin import GroupMixinLibGAP


class GroupLibGAP(GroupMixinLibGAP, Group, ParentLibGAP):

    Element = ElementLibGAP

    def __init__(self, *args, **kwds):
        """
        Group interface for LibGAP-based groups.

        INPUT:

        Same as :class:`~sage.groups.libgap_wrapper.ParentLibGAP`.

        TESTS::

            sage: F.<a,b> = FreeGroup()
            sage: G_gap = libgap.Group([ (a*b^2).gap() ])
            sage: from sage.groups.libgap_group import GroupLibGAP
            sage: G = GroupLibGAP(G_gap);  G
            Group([ a*b^2 ])
            sage: g = G.gen(0);  g
            a*b^2
            sage: TestSuite(G).run(skip=['_test_pickling', '_test_elements'])
            sage: TestSuite(g).run(skip=['_test_pickling'])
        """
        ParentLibGAP.__init__(self, *args, **kwds)
        Group.__init__(self)



import pytest

import sage.all
from sage.manifolds.differentiable.examples.symplectic_space import (
    StandardSymplecticSpace,
)
from sage.manifolds.differentiable.symplectic_form import SymplecticForm


class TestR2VectorSpace:
    @pytest.fixture
    def M(self):
        return StandardSymplecticSpace(2, 'R2', symplectic_name='omega')

    @pytest.fixture
    def omega(self, M: StandardSymplecticSpace):
        return M.symplectic_form()

    def test_repr(self, M: StandardSymplecticSpace):
        assert str(M) == "Standard symplectic space R2"

    def test_display(self, omega: SymplecticForm):
        assert str(omega.display()) == r"omega = -dq∧dp"


class TestR4VectorSpace:
    @pytest.fixture
    def M(self):
        return StandardSymplecticSpace(4, 'R4', symplectic_name='omega')

    @pytest.fixture
    def omega(self, M: StandardSymplecticSpace):
        return M.symplectic_form()

    def test_repr(self, M: StandardSymplecticSpace):
        assert str(M) == "Standard symplectic space R4"

    def test_display(self, omega: SymplecticForm):
        assert str(omega.display()) == r"omega = -dq1∧dp1 - dq2∧dp2"



r"""
Introduction to the `p`-adics
==========================================

This tutorial outlines what you need to know in order to use
`p`-adics in Sage effectively.

Our goal is to create a rich structure of different options that
will reflect the mathematical structures of the `p`-adics.
This is very much a work in progress: some of the classes that we
eventually intend to include have not yet been written, and some of
the functionality for classes in existence has not yet been
implemented. In addition, while we strive for perfect code, bugs
(both subtle and not-so-subtle) continue to evade our clutches. As
a user, you serve an important role. By writing non-trivial code
that uses the `p`-adics, you both give us insight into what
features are actually used and also expose problems in the code for
us to fix.

Our design philosophy has been to create a robust, usable interface
working first, with simple-minded implementations underneath. We
want this interface to stabilize rapidly, so that users' code does
not have to change. Once we get the framework in place, we can go
back and work on the algorithms and implementations underneath. All
of the current `p`-adic code is currently written in pure
Python, which means that it does not have the speed advantage of
compiled code. Thus our `p`-adics can be painfully slow at
times when you're doing real computations. However, finding and
fixing bugs in Python code is *far* easier than finding and fixing
errors in the compiled alternative within Sage (Cython), and Python
code is also faster and easier to write. We thus have significantly
more functionality implemented and working than we would have if we
had chosen to focus initially on speed. And at some point in the
future, we will go back and improve the speed. Any code you have
written on top of our `p`-adics will then get an immediate
performance enhancement.

If you do find bugs, have feature requests or general comments,
please email sage-support@groups.google.com or
roed@math.harvard.edu.

Terminology and types of `p`-adics
==========================================

To write down a general `p`-adic element completely would
require an infinite amount of data. Since computers do not have
infinite storage space, we must instead store finite approximations
to elements. Thus, just as in the case of floating point numbers
for representing reals, we have to store an element to a finite
precision level. The different ways of doing this account for the
different types of `p`-adics.

We can think of `p`-adics in two ways. First, as a
projective limit of finite groups:

.. MATH::

    \ZZ_p = \lim_{\leftarrow n} \ZZ/p^n\ZZ.

Secondly, as Cauchy sequences of rationals (or integers, in the
case of `\ZZ_p`) under the `p`-adic metric.
Since we only need to consider these sequences up to equivalence,
this second way of thinking of the `p`-adics is the same as
considering power series in `p` with integral coefficients
in the range `0` to `p-1`. If we only allow
nonnegative powers of `p` then these power series converge
to elements of `\ZZ_p`, and if we allow bounded
negative powers of `p` then we get `\QQ_p`.

Both of these representations give a natural way of thinking about
finite approximations to a `p`-adic element. In the first
representation, we can just stop at some point in the projective
limit, giving an element of `\ZZ/p^n\ZZ`. As
`\ZZ_p / p^n\ZZ_p \cong \ZZ/p^n\ZZ`, this
is equivalent to specifying our element modulo
`p^n\ZZ_p`.

The *absolute precision* of a finite approximation
`\bar{x} \in \ZZ/p^n\ZZ` to `x \in \ZZ_p`
is the nonnegative integer `n`.

In the second representation, we can achieve the same thing by
truncating a series

.. MATH::

      a_0 + a_1 p + a_2 p^2 + \cdots

at `p^n`, yielding

.. MATH::

     a_0 + a_1 p + \cdots + a_{n-1} p^{n-1} + O(p^n).


As above, we call this `n` the absolute precision of our
element.

Given any `x \in \QQ_p` with `x \ne 0`, we
can write `x = p^v u` where `v \in \ZZ` and
`u \in \ZZ_p^{\times}`. We could thus also store an element
of `\QQ_p` (or `\ZZ_p`) by storing
`v` and a finite approximation of `u`. This
motivates the following definition: the *relative precision* of an
approximation to `x` is defined as the absolute precision
of the approximation minus the valuation of `x`. For
example, if
`x = a_k p^k + a_{k+1} p^{k+1} +
\cdots + a_{n-1} p^{n-1} + O(p^n)`
then the absolute precision of `x` is `n`, the
valuation of `x` is `k` and the relative precision
of `x` is `n-k`.

There are three different representations of `\ZZ_p`
in Sage and one representation of `\QQ_p`:

-  the fixed modulus ring

-  the capped absolute precision ring

-  the capped relative precision ring, and

-  the capped relative precision field.

Fixed Modulus Rings
-------------------

The first, and simplest, type of `\ZZ_p` is basically
a wrapper around `\ZZ/p^n\ZZ`, providing a unified
interface with the rest of the `p`-adics. You specify a
precision, and all elements are stored to that absolute precision.
If you perform an operation that would normally lose precision, the
element does not track that it no longer has full precision.

The fixed modulus ring provides the lowest level of convenience,
but it is also the one that has the lowest computational overhead.
Once we have ironed out some bugs, the fixed modulus elements will
be those most optimized for speed.

As with all of the implementations of `\ZZ_p`, one
creates a new ring using the constructor ``Zp``, and passing in
``'fixed-mod'`` for the ``type`` parameter. For example,

::

    sage: R = Zp(5, prec = 10, type = 'fixed-mod', print_mode = 'series')
    sage: R
    5-adic Ring of fixed modulus 5^10

One can create elements as follows::

    sage: a = R(375)
    sage: a
    3*5^3
    sage: b = R(105)
    sage: b
    5 + 4*5^2

Now that we have some elements, we can do arithmetic in the ring.

::

    sage: a + b
    5 + 4*5^2 + 3*5^3
    sage: a * b
    3*5^4 + 2*5^5 + 2*5^6

Floor division (//) divides even though the result isn't really
known to the claimed precision; note that division isn't defined::

    sage: a // 5
    3*5^2

::

    sage: a / 5
    Traceback (most recent call last):
    ...
    ValueError: cannot invert non-unit

Since elements don't actually store their actual precision, one can
only divide by units::

    sage: a / 2
    4*5^3 + 2*5^4 + 2*5^5 + 2*5^6 + 2*5^7 + 2*5^8 + 2*5^9
    sage: a / b
    Traceback (most recent call last):
    ...
    ValueError: cannot invert non-unit

If you want to divide by a non-unit, do it using the ``//``
operator::

    sage: a // b
    3*5^2 + 3*5^3 + 2*5^5 + 5^6 + 4*5^7 + 2*5^8 + 3*5^9

Capped Absolute Rings
---------------------

The second type of implementation of `\ZZ_p` is
similar to the fixed modulus implementation, except that individual
elements track their known precision. The absolute precision of
each element is limited to be less than the precision cap of the
ring, even if mathematically the precision of the element would be
known to greater precision (see Appendix A for the reasons for the
existence of a precision cap).

Once again, use ``Zp`` to create a capped absolute `p`-adic
ring.

::

    sage: R = Zp(5, prec = 10, type = 'capped-abs', print_mode = 'series')
    sage: R
    5-adic Ring with capped absolute precision 10

We can do similar things as in the fixed modulus case::

    sage: a = R(375)
    sage: a
    3*5^3 + O(5^10)
    sage: b = R(105)
    sage: b
    5 + 4*5^2 + O(5^10)
    sage: a + b
    5 + 4*5^2 + 3*5^3 + O(5^10)
    sage: a * b
    3*5^4 + 2*5^5 + 2*5^6 + O(5^10)
    sage: c = a // 5
    sage: c
    3*5^2 + O(5^9)

Note that when we divided by 5, the precision of ``c`` dropped.
This lower precision is now reflected in arithmetic.

::

    sage: c + b
    5 + 2*5^2 + 5^3 + O(5^9)

Division is allowed: the element that results is a capped relative
field element, which is discussed in the next section::

    sage: 1 / (c + b)
    5^-1 + 3 + 2*5 + 5^2 + 4*5^3 + 4*5^4 + 3*5^6 + O(5^7)

Capped Relative Rings and Fields
--------------------------------

Instead of restricting the absolute precision of elements (which
doesn't make much sense when elements have negative valuations),
one can cap the relative precision of elements. This is analogous
to floating point representations of real numbers. As in the reals,
multiplication works very well: the valuations add and the relative
precision of the product is the minimum of the relative precisions
of the inputs. Addition, however, faces similar issues as floating
point addition: relative precision is lost when lower order terms
cancel.

To create a capped relative precision ring, use ``Zp`` as before.
To create capped relative precision fields, use ``Qp``.

::

    sage: R = Zp(5, prec = 10, type = 'capped-rel', print_mode = 'series')
    sage: R
    5-adic Ring with capped relative precision 10
    sage: K = Qp(5, prec = 10, type = 'capped-rel', print_mode = 'series')
    sage: K
    5-adic Field with capped relative precision 10

We can do all of the same operations as in the other two cases, but
precision works a bit differently: the maximum precision of an
element is limited by the precision cap of the ring.

::

    sage: a = R(375)
    sage: a
    3*5^3 + O(5^13)
    sage: b = K(105)
    sage: b
    5 + 4*5^2 + O(5^11)
    sage: a + b
    5 + 4*5^2 + 3*5^3 + O(5^11)
    sage: a * b
    3*5^4 + 2*5^5 + 2*5^6 + O(5^14)
    sage: c = a // 5
    sage: c
    3*5^2 + O(5^12)
    sage: c + 1
    1 + 3*5^2 + O(5^10)

As with the capped absolute precision rings, we can divide,
yielding a capped relative precision field element.

::

    sage: 1 / (c + b)
    5^-1 + 3 + 2*5 + 5^2 + 4*5^3 + 4*5^4 + 3*5^6 + 2*5^7 + 5^8 + O(5^9)

Unramified Extensions
---------------------

One can create unramified extensions of `\ZZ_p` and
`\QQ_p` using the functions ``Zq`` and ``Qq``.

In addition to requiring a prime power as the first argument,
``Zq`` also requires a name for the generator of the residue field.
One can specify this name as follows::

    sage: R.<c> = Zq(125, prec=20); R                                                   # needs sage.libs.ntl
    5-adic Unramified Extension Ring in c defined by x^3 + 3*x + 3

Eisenstein Extensions
---------------------

It is also possible to create Eisenstein extensions of `\ZZ_p`
and `\QQ_p`.  In order to do so, create the ground field first::

    sage: R = Zp(5, 2)

Then define the polynomial yielding the desired extension.::

    sage: S.<x> = ZZ[]
    sage: f = x^5 - 25*x^3 + 15*x - 5

Finally, use the ``ext`` function on the ground field to create the
desired extension.::

    sage: W.<w> = R.ext(f)                                                              # needs sage.libs.ntl sage.rings.padics

You can do arithmetic in this Eisenstein extension::

    sage: (1 + w)^7                                                                     # needs sage.libs.ntl sage.rings.padics
    1 + 2*w + w^2 + w^5 + 3*w^6 + 3*w^7 + 3*w^8 + w^9 + O(w^10)

Note that the precision cap increased by a factor of 5, since the
ramification index of this extension over `\ZZ_p` is 5.
"""
